# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import pytest

import pylife.strength.miner as miner
import pylife.stress

from . import data


def make_collective_from_raw_data(raw_collective):
    smallest_length = np.abs(np.diff(raw_collective[:, 0])).min()
    amplitude_left = raw_collective[:, 0] - smallest_length/2.
    amplitude_right = raw_collective[:, 0] + smallest_length/2

    decumulated = np.append(np.abs(np.diff(raw_collective[:, 1])), raw_collective[-1, 1])
    coll = pd.Series(decumulated,
                     index=pd.IntervalIndex.from_arrays(2.*amplitude_left, 2.*amplitude_right, name='range'),
                     name='cycles')
    return coll


@pytest.fixture
def collective(request):
    request.cls.coll = make_collective_from_raw_data(data.collective).load_collective


# parameters for the example given in Haibach2006
sn_curve_parameters_haibach = pd.Series({
    "ND": 1E6,
    "k_1": 4,
    "SD": 100
})
# parameters for the example given in Haibach2006
sn_curve_parameters_elementary = pd.Series({
    "ND": 2E6,
    "k_1": 7,
    "SD": 125
})


@pytest.fixture
def miner_elementary():
    return miner.MinerElementary.from_parameters(ND=1e6, k_1=6, SD=200.)


@pytest.fixture
def miner_haibach():
    return miner.MinerHaibach.from_parameters(ND=10**6, k_1=6, SD=200)


def test_effective_damage_sum_limitations():
    # lower limit for d_m = 1
    assert miner.effective_damage_sum(16) == 1.0
    # upper limit for d_m = 0.3
    assert miner.effective_damage_sum(1975.308641975309) == 0.3


@pytest.mark.usefixtures('collective')
class TestMinerElementary():

    def test_lifetime_factor(self, miner_elementary):
        z = 1.3180413239445  # = (ND / H0)**(1. / k)
        np.testing.assert_almost_equal(miner_elementary.finite_life_factor(190733.0), z)

    def test_lifetime_multiple(self, miner_elementary):
        # test requires k = 6

        A = miner_elementary.lifetime_multiple(self.coll)
        np.testing.assert_almost_equal(A, 862.99608075)

    def test_d_m(self, miner_elementary):
        d_m = miner_elementary.effective_damage_sum(self.coll)
        d_m_test = 0.36900120961677296
        np.testing.assert_almost_equal(d_m, d_m_test)

    def test_N_predict(self, miner_elementary):
        load_level = 400
        expected_cycles = 13484313.761758052
        coll = self.coll.scale(load_level/self.coll.amplitude.max())
        result = miner_elementary.gassner_cycles(coll)
        np.testing.assert_almost_equal(result, expected_cycles)

    def test_gassner(self, miner_elementary):
        load_level = 400
        expected_cycles = 13484313.761758052
        foo_collective = pd.Series({
            'amplitude': load_level,
            'cycles': expected_cycles
        })

        gassner = miner_elementary.gassner(self.coll)
        expected_damage = pd.Series(1.0, name='damage')
        pd.testing.assert_series_equal(gassner.damage(foo_collective), expected_damage)

    def test_damage_accumulation_validation(self):
        """The test uses data from the book Haibach2006

        See miner.py for reference.
        The examples can be found on page 271.
        """
        coll = make_collective_from_raw_data(data.coll_elementary_acc)
        load_level = coll.load_collective.amplitude.max()
        expected_N = 2167330
        miner_elementary = sn_curve_parameters_elementary.gassner_miner_elementary
        # some rounding is assumed in the example from the book
        # so in respect to millions of cycles a small neglectable tolerance is accepted
        gassner = miner_elementary.gassner(coll.load_collective)
        np.testing.assert_approx_equal(expected_N, gassner.cycles(load_level), significant=6)


@pytest.mark.usefixtures('collective')
class TestMinerHaibach:

    def test_lifetime_multiple_split_damage_regions(self, miner_haibach):
        load_level = 400
        coll = self.coll.scale(load_level/self.coll.amplitude.max())
        miner_haibach.lifetime_multiple(coll)

    def test_lifetime_multiple(self, miner_haibach):
        load_level = 400
        coll = self.coll.scale(load_level/self.coll.amplitude.max())
        A = miner_haibach.lifetime_multiple(coll)
        np.testing.assert_almost_equal(A, 1000.342377197)

    def test_lifetime_multiple_no_load_level(self, miner_haibach):
        A = miner_haibach.lifetime_multiple(self.coll.scale(200.))
        np.testing.assert_almost_equal(A, 1061.21644181784)

    def test_N_predict(self, miner_haibach):
        load_level = 400
        coll = self.coll.scale(load_level/self.coll.amplitude.max())
        N_woehler_load_level = (
            miner_haibach.ND * (load_level / miner_haibach.SD)**(-miner_haibach.k_1)
        )
        A = miner_haibach.lifetime_multiple(coll)
        N_predict = N_woehler_load_level * A

        coll = self.coll.scale(load_level/self.coll.amplitude.max())

        np.testing.assert_almost_equal(miner_haibach.gassner_cycles(coll), N_predict)

    @pytest.mark.parametrize("load_level, predicted_N", [
        (100, 935519000),
        (150, 61605400),
        (200, 11994100),
        (300, 1761820),
        (400, 505981),
        (600, 99382),
        (800, 31103)
    ])
    def test_damage_accumulation_validation(self, load_level, predicted_N):
        """The test uses data from the book Haibach2006

        See miner.py for reference.
        The examples can be found on page 292.
        """
        coll = make_collective_from_raw_data(data.coll_haibach_mod_acc)
        coll = coll.load_collective.scale(load_level/coll.load_collective.amplitude.max())

        miner_haibach = sn_curve_parameters_haibach.gassner_miner_haibach
        # some rounding is assumed in the example from the book
        # so in respect to millions of cycles a small neglectable tolerance is accepted
        np.testing.assert_approx_equal(miner_haibach.gassner_cycles(coll), predicted_N, significant=5)



import pylife.stress.rainflow as RF


def test_threepoint(benchmarker):
    benchmark_time = 0.05  # seconds
    elapsed = benchmarker(RF.ThreePointDetector(recorder=RF.FullRecorder()))

    assert (
        elapsed < benchmark_time
    ), f"Benchmark time of {benchmark_time} s not exceeded. Needed {elapsed:0.4f} s."


def test_fourpoint(benchmarker):
    benchmark_time = 0.05  # seconds
    elapsed = benchmarker(RF.FourPointDetector(recorder=RF.FullRecorder()))

    assert (
        elapsed < benchmark_time
    ), f"Benchmark time of {benchmark_time} s not exceeded. Needed {elapsed:0.4f} s."





cimport cython
import numpy as np
from libc.math cimport fabs


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
def fourpoint_loop(double [::1] turns, size_t [::1] turns_index):
    cdef Py_ssize_t len_turns = len(turns)

    from_vals = np.empty(len_turns//2, dtype=np.float64)
    to_vals = np.empty(len_turns//2, dtype=np.float64)
    from_index = np.empty(len_turns//2, dtype=np.uintp)
    to_index = np.empty(len_turns//2, dtype=np.uintp)

    cdef double [::1] from_vals_v = from_vals
    cdef double [::1] to_vals_v = to_vals
    cdef size_t [::1] from_index_v = from_index
    cdef size_t [::1] to_index_v = to_index

    residual_index = np.empty(len_turns, dtype=np.uintp)
    cdef size_t [::1] residual_index_v = residual_index

    residual_index_v[0] = 0
    residual_index_v[1] = 1

    cdef size_t i = 2
    cdef size_t ri = 2
    cdef size_t t = 0

    cdef double a
    cdef double b
    cdef double c
    cdef double d
    cdef double ab
    cdef double bc
    cdef double cd

    while i < len_turns:
        if ri < 3:
            residual_index_v[ri] = i
            ri += 1
            i += 1
            continue

        a = turns[residual_index_v[ri-3]]
        b = turns[residual_index_v[ri-2]]
        c = turns[residual_index_v[ri-1]]
        d = turns[i]

        ab = fabs(a - b)
        bc = fabs(b - c)
        cd = fabs(c - d)
        if bc <= ab and bc <= cd:
            from_vals_v[t] = b
            to_vals_v[t] = c

            ri -= 1
            to_index_v[t] = turns_index[residual_index_v[ri]]
            ri -= 1
            from_index_v[t] = turns_index[residual_index_v[ri]]
            t += 1
            continue

        residual_index_v[ri] = i
        ri += 1
        i += 1

    return from_vals[:t], to_vals[:t], from_index[:t], to_index[:t], residual_index[:ri]


cpdef double _max(double a, double b):
    return a if a > b else b


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
def threepoint_loop(double [::1] turns, size_t [::1] turns_index, size_t highest_front,
                    size_t lowest_front, size_t residual_length):
    cdef Py_ssize_t len_turns = len(turns)

    from_vals = np.empty(len_turns//2, dtype=np.float64)
    to_vals = np.empty(len_turns//2, dtype=np.float64)
    from_index = np.empty(len_turns//2, dtype=np.uintp)
    to_index = np.empty(len_turns//2, dtype=np.uintp)

    cdef double [::1] from_vals_v = from_vals
    cdef double [::1] to_vals_v = to_vals
    cdef size_t [::1] from_index_v = from_index
    cdef size_t [::1] to_index_v = to_index

    residual_index = np.empty(len_turns, dtype=np.uintp)
    residual_index[:residual_length] = np.arange(residual_length)
    cdef size_t [::1] residual_index_v = residual_index

    residual_index_v[0] = 0
    residual_index_v[1] = 1

    cdef size_t ri = 2
    cdef size_t t = 0

    cdef size_t back = residual_index_v[1] + 1
    cdef size_t front
    cdef size_t start

    cdef double start_val
    cdef double front_val
    cdef double back_valstar

    while back < len_turns:
        if ri >= 2:
            start = residual_index_v[ri-2]
            front = residual_index_v[ri-1]
            start_val, front_val, back_val = turns[start], turns[front], turns[back]

            if front_val > turns[highest_front]:
                highest_front = front
            elif front_val < turns[lowest_front]:
                lowest_front = front
            elif (start >= _max(lowest_front, highest_front) and
                  fabs(back_val - front_val) >= fabs(front_val - start_val)):
                from_vals_v[t] = start_val
                to_vals_v[t] = front_val

                from_index_v[t] = turns_index[start]
                to_index_v[t] = turns_index[front]

                t += 1
                ri -= 2
                continue

        residual_index[ri] = back
        ri += 1
        back += 1

    return from_vals[:t], to_vals[:t], from_index[:t], to_index[:t], residual_index[:ri]



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys, os, copy
import warnings
import pytest

import numpy as np
import pandas as pd
import numpy.testing as testing

import pylife.strength.meanstress as MST


@pytest.fixture
def ms_sens():
    return pd.Series({'M': 0.5, 'M2': 0.5/3.})


def test_meanstress_collective_empty_fail():
    df = pd.Series([], index=pd.MultiIndex.from_tuples([], names= ['foo', 'bar']), dtype=np.float64)

    with pytest.raises(AttributeError, match="Load collective"):
        df.meanstress_transform


def test_meanstress_collective_empty_fkm_goodman(ms_sens):
    from_intervals = pd.interval_range(0., 0., 0)
    to_intervals = pd.interval_range(0., 0., 0)
    index = pd.MultiIndex.from_product([from_intervals, to_intervals], names=['from', 'to'])

    ser = pd.Series([], index=index, dtype=np.float64)

    expected = pd.Series([],
                         index=pd.IntervalIndex(to_intervals, name='range'),
                         name='cycles', dtype=np.float64)
    res = ser.meanstress_transform.fkm_goodman(ms_sens, -1.).to_pandas()
    pd.testing.assert_series_equal(res, expected)


def test_meanstress_collective_fkm_goodman_single_ms_sens(ms_sens):
    fr = pd.IntervalIndex.from_breaks(np.linspace(-25./24., 1., 49), closed='left')
    to = pd.IntervalIndex.from_breaks(np.linspace(-1./24., 2., 49), closed='left')
    index = pd.MultiIndex.from_product([fr, to], names=['from', 'to'])

    rf = pd.Series(0.0, name='cycles', index=index)

    rf.loc[(14./24., 21./12.)] = 1
    rf.loc[(0., 4./3.)] = 3
    rf.loc[(-1., 1.)] = 5

    haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
    R_goal = -1.

    expected = 2.0
    expected_interval = pd.Interval(expected - 1./96., expected + 1./96.)
    res = rf.meanstress_transform.fkm_goodman(haigh, R_goal).to_pandas()

    mask = res.index.get_level_values('range').overlaps(expected_interval)
    assert res.loc[mask].sum() == 9
    assert res.loc[~mask].sum() == 0


def test_meanstress_collective_fkm_goodman_multiple_ms_sens():
    df = pd.DataFrame({
        'from': [-6., -4.,  -5./2., -1., -0.4, 0., 7./12.],
        'to': [-2., 0., 0.5, 1., 1.2, 4./3., 21./12.]
    }, index=pd.Index([3, 4, 5, 6, 7, 8, 9], name='element_id'))

    ms_sens = pd.DataFrame({
        'M': [0.5, 0.4],
        'M2': [0.5/3., 0.4/3.]
    })

    expected_index = pd.MultiIndex.from_tuples([
        (0, 3), (0, 4), (0, 5), (0, 6), (0, 7), (0, 8), (0, 9),
        (1, 3), (1, 4), (1, 5), (1, 6), (1, 7), (1, 8), (1, 9)
    ], names=[None, 'element_id'])

    res = df.meanstress_transform.fkm_goodman(ms_sens, -1.)

    expected_amplitude = pd.Series(
        [1., 1., 1., 1., 1., 1., 1., 1.2, 1.2, 1.1, 1.0, 0.96, 0.93, 0.91],
        index=expected_index,
        name='amplitude'
    )
    pd.testing.assert_series_equal(res.amplitude, expected_amplitude, rtol=1e-2)

    expected_meanstress = pd.Series(
        np.zeros(14),
        index=expected_index,
        name='meanstress'
    )
    pd.testing.assert_series_equal(res.meanstress, expected_meanstress, rtol=1e-2)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import sys
import importlib
import pytest
import pandas as pd
import numpy as np
import unittest.mock as mock

from .data import *

try:
    import pymc
    import bambi
    HAVE_PYMC_AND_BAMBI = True
except ModuleNotFoundError:
    HAVE_PYMC_AND_BAMBI = False

from pylife.materialdata import woehler


pytestmark = pytest.mark.skip("Bayesian Wöhler estimator is no longer in use")


@pytest.fixture
def core_num():
    return 1 if sys.platform.startswith('win') else None


@pytest.mark.skipif(not HAVE_PYMC_AND_BAMBI, reason="Don't have pymc")
@mock.patch('pylife.materialdata.woehler.bayesian.pm')
@mock.patch('pylife.materialdata.woehler.bayesian.bambi.Model')
def test_bayesian_slope_trace(bambi_model, pm, core_num):
    bambi_model_object = mock.MagicMock()
    bambi_model.return_value = bambi_model_object
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    bayes = woehler.Bayesian(fd)
    bayes._nsamples = 1000
    bayes._slope_trace()

    formula, data_dict = bambi_model.call_args[0]
    assert formula == 'y ~ x'
    np.testing.assert_array_equal(data_dict['x'], np.log10(fd.fractures.load))
    np.testing.assert_array_equal(data_dict['y'], np.log10(fd.fractures.cycles.to_numpy()))
    assert bambi_model.call_args[1]['family'] == 't'  # Consider switch to kwargs property when py3.7 is dropped

    bambi_model_object.fit.assert_called_with(draws=1000, cores=core_num, tune=1000, chains=2, random_seed=None)


@pytest.mark.skipif(not HAVE_PYMC_AND_BAMBI, reason="Don't have pymc")
@mock.patch('pylife.materialdata.woehler.bayesian.pm')
def test_bayesian_TN_trace(pm, core_num):
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    bayes = woehler.Bayesian(fd)
    bayes._common_analysis()
    bayes._nsamples = 1000
    bayes._TN_trace()

    pm.HalfNormal.assert_called_with('stdev', sigma=1.3)

    assert pm.Normal.call_count == 2

    expected_mu = 5.294264482012933
    expected_sigma = 0.2621494419382026

    assert pm.Normal.call_args_list[0][0] == ('mu',)
    np.testing.assert_almost_equal(pm.Normal.call_args_list[0][1]['mu'], expected_mu, decimal=9)
    np.testing.assert_almost_equal(pm.Normal.call_args_list[0][1]['sigma'], expected_sigma, decimal=9)

    assert pm.Normal.call_args_list[1][0] == ('y',)
    observed = pm.Normal.call_args_list[1][1]['observed']  # Consider switch to kwargs property when py3.7 is dropped
    np.testing.assert_almost_equal(observed.mean(), expected_mu, decimal=9)
    np.testing.assert_almost_equal(observed.std(), expected_sigma, decimal=9)

    pm.sample.assert_called_with(1000, cores=core_num, target_accept=0.99, random_seed=None, chains=3, tune=1000)


@pytest.mark.skipif(not HAVE_PYMC_AND_BAMBI, reason="Don't have pymc")
@mock.patch('pylife.materialdata.woehler.bayesian.pt')
@mock.patch('pylife.materialdata.woehler.bayesian.pm')
def test_bayesian_SD_TS_trace_mock(pm, pt, core_num):
    def check_likelihood(l, var):
        assert var == pt.as_tensor_variable.return_value
        assert isinstance(l.likelihood, woehler.likelihood.Likelihood)
        np.testing.assert_array_equal(l.likelihood._fd, fd)
        return 'foovar'

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    inf_load_mean = fd.infinite_zone.load.mean()
    inf_load_std = fd.infinite_zone.load.std()

    with mock.patch.object(woehler.Bayesian._LogLike, '__call__', autospec=True) as loglike_call:
        loglike_call.side_effect = check_likelihood

        bayes = woehler.Bayesian(fd)
        bayes._nsamples = 1000
        bayes._SD_TS_trace()

    pm.Normal.assert_called_once_with('SD', mu=inf_load_mean, sigma=inf_load_std * 5)
    pm.Lognormal.assert_called_once()
    np.testing.assert_approx_equal(pm.Lognormal.call_args_list[0][1]['mu'], np.log10(1.1))
    np.testing.assert_approx_equal(pm.Lognormal.call_args_list[0][1]['sigma'], 0.3)

    pt.as_tensor_variable.assert_called_once_with([pm.Normal.return_value, pm.Lognormal.return_value])

    pm.Potential.assert_called_once_with('likelihood', 'foovar')

    pm.sample.assert_called_with(
        1000, cores=core_num, chains=3, random_seed=None, discard_tuned_samples=True, tune=1000
    )


@pytest.mark.skipif(not HAVE_PYMC_AND_BAMBI, reason="Don't have pymc")
@mock.patch('pylife.materialdata.woehler.bayesian.Bayesian._SD_TS_trace')
@mock.patch('pylife.materialdata.woehler.bayesian.Bayesian._TN_trace')
@mock.patch('pylife.materialdata.woehler.bayesian.Bayesian._slope_trace')
def test_bayesian_mock(_slope_trace, _TN_trace, _SD_TS_trace):
    import xarray as xr
    expected = pd.Series({
        'SD': 100.,
        'TS': 1.12,
        'k_1': 7.0,
        'ND': 1e6,
        'TN': 5.3,
        'failure_probability': 0.5
    }).sort_index()

    expected_slope_trace = xr.Dataset({
        'x': ([0, 1], np.array([[0.0, -8.0, -6.0], [0.0, -8.0, -6.0]])),
        'Intercept': ([0, 1], np.array([[0.0, 19., 21.], [0.0, 19., 21.]]))
    })

    expected_SD_TS_trace = xr.Dataset({
        'SD': ([0, 1], np.array([[0.0, 150., 50], [0.0, 150., 50]])),
        'TS': ([0, 1], np.array([[0.0, 1.22, 1.02], [0.0, 1.22, 1.02]]))
    })

    expected_TN_trace = xr.Dataset({
        'mu': ([0, 1], np.array([[0.0, 5.4, 5.2], [0.0, 5.4, 5.2]]))
    })

    _slope_trace.return_value = expected_slope_trace
    _TN_trace.return_value = expected_TN_trace
    _SD_TS_trace.return_value = expected_SD_TS_trace

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.Bayesian(fd).analyze(nsamples=10).sort_index()

    pd.testing.assert_series_equal(wc, expected)


@pytest.mark.skipif(not HAVE_PYMC_AND_BAMBI, reason="Don't have pymc")
#@pytest.mark.slow_acceptance
def test_bayesian_full():
    expected = pd.Series({
        'SD': 340.,
        'TS': 1.12,
        'k_1': 7.0,
        'ND': 400000.,
        'TN': 5.3,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.Bayesian(fd).analyze(random_seed=4223, progressbar=False).sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__


import numpy as np
import pandas as pd


class Broadcaster:
    """The Broadcaster to align pyLife signals to operands.

    Parameters
    ----------
    pandas_obj : :class:`pandas.Series` or :class:`pandas.DataFrame`
       the object of the ``Broadcaster``


    In most cases the ``Broadcaster`` class is not used directly.  The
    functionality is in most cases used by the derived class
    :class:`~pylife.PylifeSignal`.

    The purpose of the ``Broadcaster`` is to take two numerical objects and
    return two objects of the same numerical data with an aligned index.  That
    means that mathematical operations using the two objects as operands can be
    implemented using numpy's broadcasting functionality.

    See method :meth:`~pylife.Broadcaster.broadcast` documentation for details.

    The broadcasting is done in the following ways:

    ::

        object                 parameter              returned object         returned parameter

        Series                 Scalar                 Series                  Scalar
        |------|-----|                                |------|-----|
        | idx  |     |                                | idx  |     |
        |------|-----|         5.0               ->   |------|-----|          5.0
        | foo  | 1.0 |                                | foo  | 1.0 |
        | bar  | 2.0 |                                | bar  | 2.0 |
        |------|-----|                                |------|-----|


        DataFrame              Scalar                 DataFrame               Series
        |------|-----|-----|                          |------|-----|-----|    |------|-----|
        | idx  | foo | bar |                          | idx  | foo | bar |    | idx  |     |
        |------|-----|-----|                          |------|-----|-----|    |------|-----|
        | 0    | 1.0 | 2.0 |   5.0               ->   | 0    | 1.0 | 2.0 |    | 0    | 5.0 |
        | 1    | 1.0 | 2.0 |                          | 1    | 1.0 | 2.0 |    | 1    | 5.0 |
        | ...  | ... | ... |                          | ...  | ... | ... |    | ...  | ... |
        |------|-----|-----|                          |------|-----|-----|    |------|-----|


        Series                 Series/DataFrame       DataFrame               Series/DataFrame
        |------|-----|         |------|-----|         |------|-----|-----|    |------|-----|
        | None |     |         | idx  |     |         | idx  | foo | bar |    | idx  |     |
        |------|-----|         |------|-----|    ->   |------|-----|-----|    |------|-----|
        | foo  | 1.0 |         | 0    | 5.0 |         | 0    | 1.0 | 2.0 |    | 0    | 5.0 |
        | bar  | 2.0 |         | 1    | 6.0 |         | 1    | 1.0 | 2.0 |    | 1    | 6.0 |
        |------|-----|         | ...  | ... |         | ...  | ... | ... |    | ...  | ... |
                               |------|-----|         |------|-----|-----|    |------|-----|


        Series/DataFrame       Series/DataFrame       Series/DataFrame        Series/DataFrame
        |------|-----|         |------|-----|         |------|-----|          |------|-----|
        | xidx |     |         | xidx |     |         | xidx |     |          | xidx |     |
        |------|-----|         |------|-----|    ->   |------|-----|          |------|-----|
        | foo  | 1.0 |         | tau  | 5.0 |         | foo  | 1.0 |          | foo  | nan |
        | bar  | 2.0 |         | bar  | 6.0 |         | bar  | 2.0 |          | bar  | 6.0 |
        |------|-----|         |------|-----|         | tau  | nan |          | tau  | 5.0 |
                                                      |------|-----|          |------|-----|


        Series/DataFrame       Series/DataFrame       Series/DataFrame        Series/DataFrame
        |------|-----|         |------|-----|         |------|------|-----|   |------|------|-----|
        | xidx |     |         | yidx |     |         | xidx | yidx |     |   | xidx | yidx |     |
        |------|-----|         |------|-----|   ->    |------|------|-----|   |------|------|-----|
        | foo  | 1.0 |         | tau  | 5.0 |         | foo  | tau  | 1.0 |   | foo  | tau  | 5.0 |
        | bar  | 2.0 |         | chi  | 6.0 |         |      | chi  | 1.0 |   |      | chi  | 6.0 |
        |------|-----|         |------|-----|         | bar  | tau  | 2.0 |   | bar  | tau  | 5.0 |
                                                      |      | chi  | 2.0 |   |      | chi  | 6.0 |
                                                      |------|------|-----|   |------|------|-----|


    """

    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def broadcast(self, parameter, droplevel=None):
        """Broadcast the parameter to the object of ``self``.

        Parameters
        ----------

        parameters : scalar, numpy array or pandas object
            The parameter to broadcast to


        Returns
        -------
        parameter, object : index aligned numerical objects


        Examples
        --------

        The behavior of the Broadcaster is best illustrated by examples:

        .. jupyter-execute::
           :hide-code:

           import pandas as pd
           from pylife import Broadcaster

        * Broadcasting :class:`pandas.Series` to a scalar results in a scalar
          and a :class:`pandas.Series`.

          .. jupyter-execute::

              obj = pd.Series([1.0, 2.0], index=pd.Index(['foo', 'bar'], name='idx'))
              obj

          .. jupyter-execute::

              parameter, obj = Broadcaster(obj).broadcast(5.0)

              parameter

          .. jupyter-execute::

              obj


        * Broadcasting :class:`pandas.DataFrame` to a scalar results in a
          :class:`pandas.DataFrame` and a :class:`pandas.Series`.

          .. jupyter-execute::

              obj = pd.DataFrame({
                  'foo': [1.0, 2.0],
                  'bar': [3.0, 4.0]
              }, index=pd.Index([1, 2], name='idx'))
              obj

          .. jupyter-execute::

              parameter, obj = Broadcaster(obj).broadcast(5.0)

              parameter

          .. jupyter-execute::

              obj


        * Broadcasting :class:`pandas.DataFrame` to a a :class:`pandas.Series`
          results in a :class:`pandas.DataFrame` and a :class:`pandas.Series`,
          **if and only if** the index name of the object is ``None``.

          .. jupyter-execute::

              obj = pd.Series([1.0, 2.0], index=pd.Index(['tau', 'chi']))
              obj

          .. jupyter-execute::

              parameter = pd.Series([3.0, 4.0], index=pd.Index(['foo', 'bar'], name='idx'))
              parameter

          .. jupyter-execute::

              parameter, obj = Broadcaster(obj).broadcast(parameter)

              parameter

          .. jupyter-execute::

              obj

        """
        droplevel = droplevel or []

        if not isinstance(parameter, pd.Series) and not isinstance(parameter, pd.DataFrame):
            if isinstance(self._obj, pd.Series):
                return self._broadcast_series(parameter)
            return self._broadcast_frame(parameter)

        if self._obj.index.names == [None] and isinstance(self._obj, pd.Series):
            df = pd.DataFrame(index=parameter.index, columns=self._obj.index)
            for c in self._obj.index:
                df[c] = self._obj[c]
            return parameter, df

        return self._broadcast_frame_to_frame(parameter, droplevel)

    def _broadcast_series(self, parameter):
        prm = np.asarray(parameter)
        if prm.shape == ():
            return prm, self._obj

        df = self._broadcasted_dataframe(parameter)
        if isinstance(parameter, pd.Series):
            return parameter, df.set_index(parameter.index, inplace=True)

        return pd.Series(prm), df

    def _broadcast_series_to_frame(self, parameter):
        return parameter, self._broadcasted_dataframe(parameter).set_index(parameter.index)

    def _broadcast_frame_to_frame(self, parameter, droplevel):
        def align_and_reorder():
            if isinstance(self._obj, pd.DataFrame) and isinstance(parameter, pd.Series):
                obj, prm = self._obj.align(pd.DataFrame({0: parameter}), axis=0)
                prm = prm.iloc[:, 0]
                prm.name = parameter.name
            else:
                obj, prm = self._obj.align(parameter, axis=0)

            if len(droplevel) > 0:
                prm_columns = list(filter(lambda level: level not in droplevel, total_columns))
                prm = prm.groupby(prm_columns).first()
            else:
                prm_columns = total_columns

            if obj.index.nlevels > 2:
                prm = prm.reorder_levels(prm_columns)
                obj = obj.reorder_levels(total_columns)
            return prm, obj

        def cross_join_and_align_obj_and_parameter():
            prm_index = parameter.index.to_frame().reset_index(drop=True)
            obj_index = self._obj.index.to_frame()[obj_index_names]
            new_index = (obj_index
                         .join(prm_index, how='cross')
                         .set_index(total_columns)
                         .reorder_levels(total_columns).index)

            obj = _broadcast_to(self._obj, new_index)
            prm = _broadcast_to(parameter, new_index)

            obj, prm = obj.align(prm, axis=0)

            if len(droplevel) > 0:
                prm_columns = list(filter(lambda level: level not in droplevel, total_columns))
                prm = prm.groupby(prm_columns).first()

            return obj, prm

        uuids = _replace_none_index_names_with_unique_string([parameter, self._obj])

        index_level_cache = _IndexLevelCache(self._obj, parameter)

        prm_index_names = list(parameter.index.names)
        obj_index_names = list(self._obj.index.names)

        total_columns = obj_index_names + [lv for lv in prm_index_names if lv not in obj_index_names]
        have_commons = len(total_columns) < len(prm_index_names) + len(obj_index_names)

        if have_commons:
            prm, obj = align_and_reorder()
        else:
            obj, prm = cross_join_and_align_obj_and_parameter()

        obj.index = index_level_cache.restore_real_index(obj.index)
        prm.index = index_level_cache.restore_real_index(prm.index)

        index_level_cache.restore_original_indeces()
        _replace_unique_string_with_none_name([obj, prm, self._obj, parameter], uuids)

        return prm, obj

    def _broadcast_frame(self, parameter):
        try:
            parameter = np.broadcast_to(parameter, len(self._obj))
        except ValueError:
            raise ValueError("Dimension mismatch. "
                             "Cannot map %d value array-like to a %d element DataFrame signal."
                             %(len(parameter), len(self._obj)))
        return pd.Series(parameter, index=self._obj.index), self._obj

    def _broadcasted_dataframe(self, parameter):
        data = np.empty((len(parameter), len(self._obj)))
        df = pd.DataFrame(data, columns=self._obj.index).assign(**self._obj)
        return df


def _broadcast_to(obj, new_index):
    if isinstance(obj, pd.DataFrame):
        new = obj
    else:
        new = pd.DataFrame(obj)

    new = pd.DataFrame(index=new_index).join(new, how='left')
    if isinstance(new_index, pd.MultiIndex):
        new = new.reorder_levels(new_index.names)
    if isinstance(obj, pd.Series):
        new = new.iloc[:, 0]
        new.name = obj.name
    return new


def _replace_none_index_names_with_unique_string(objs):
    import uuid

    def make_uuid():
        this_uuid = uuid.uuid4().hex
        uuids.append(this_uuid)
        return this_uuid

    uuids = []

    for obj in objs:
        obj.index.names = [name if name is not None else make_uuid() for name in obj.index.names]

    return uuids


def _replace_unique_string_with_none_name(objs, uuids):
    for obj in objs:
        obj.index.names = [None if name in uuids else name for name in obj.index.names]


class _IndexLevelCache:

    def __init__(self, obj, operand):

        self._obj_index = obj.index
        self._operand_index = operand.index

        self._obj = obj
        self._operand = operand

        common = set(obj.index.names).intersection(operand.index.names)
        only_obj = set(obj.index.names).difference(common)
        only_operand = set(operand.index.names).difference(common)

        self.index_levels = {}

        for name in common:
            obj_level = obj.index.get_level_values(name)
            operand_level = operand.index.get_level_values(name)
            self.index_levels[name] = obj_level.append(operand_level).unique()

        for name in only_obj:
            self.index_levels[name] = obj.index.get_level_values(name).unique()

        for name in only_operand:
            self.index_levels[name] = operand.index.get_level_values(name).unique()

        self.new_index_obj = self._make_new_index(obj.index)
        self.new_index_operand = self._make_new_index(operand.index)

        obj.index = self.new_index_obj
        operand.index = self.new_index_operand

    def restore_original_indeces(self):
        self._obj.index = self._obj_index
        self._operand.index =self._operand_index

    def restore_real_index(self, new_index):

        if len(new_index.names) > 1:
            real_index = pd.MultiIndex.from_arrays(
                [
                    self.index_levels[name][new_index.get_level_values(name)]
                    for name in new_index.names
                ],
                names=new_index.names
            )
        else:
            real_index = pd.Index(self.index_levels[new_index.name][new_index], name=new_index.name)

        return real_index

    def _make_new_index(self, index):
        if len(index.names) == 1:
            return pd.Index(self.index_levels[index.name].get_indexer_for(index), name=index.name)

        return pd.MultiIndex.from_arrays(
            [
                self.index_levels[name].get_indexer_for(index.get_level_values(name))
                for name in index.names
            ],
            names=index.names
        )



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

Equivalent Stresses
===================

Library to calculate the equivalent stress values of a FEM stress tensor.

By now the following calculation methods are implemented:

* Principal stresses
* Maximum principal stress
* Minimum principal stress
* Absolute maximum principal stress
* Von Mises
* Signed von Mises, sign from trace
* Signed von Mises, sign from absolute maximum principal stress
* Tresca
* Signed Tresca, sign from trace
* Signed Tresca, sign from absolute maximum principal stress

"""

__author__ = "Johannes Mueller, Vivien Le Baube et. al."
__maintainer__ = "Johannes Mueller"

import numpy as np
import pandas as pd
from pylife.stress import stresssignal


def eigenval(s11, s22, s33, s12, s13, s23):
    """Calculate eigenvalues of a symmetric 3D tensor.

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Array containing eigenvalues sorted in ascending order.
        Shape is (length of components, 3) or simply 3 if components are single
        values.
    """
    a = np.array([[s11, s12, s13],
                  [s12, s22, s23],
                  [s13, s23, s33]]).T
    return np.linalg.eigvalsh(a)


def _sign_trace(s11, s22, s33):
    """Calculate sign of trace. Sign of 0 is set to 1.

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Array containing sign of trace. Shape is the same as the components.
    """
    s11 = np.array(s11)
    s22 = np.array(s22)
    s33 = np.array(s33)
    assert (s11.shape == s22.shape and
            s11.shape == s33.shape), "Components' shape is not consistent."
    sgn = np.sign(s11 + s22 + s33)  # calculate sign of trace, careful: sign of 0 is 0
    if sgn.ndim == 0:
        if sgn == 0:
            sgn = np.array(1)
    else:
        sgn[sgn == 0] = 1
    return sgn


def _sign_abs_max_principal(s11, s22, s33, s12, s13, s23):
    """Calculate sign of absolute maximum principal stress. Sign of 0 is set to
     1.

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Array containing sign of absolute maximum principal stress. Shape is the
        same as the components.
    """
    w = eigenval(s11, s22, s33, s12, s13, s23).T
    w_max = np.amax(w, axis=0)
    w_min = np.amin(w, axis=0)
    sgn = np.sign(w_max + w_min)
    # sign of 0 is 0, replace it with 1
    zero_sign_bool = np.array(sgn == 0, dtype=int)
    sgn = sgn + zero_sign_bool  # works for all dimensions, no need to differentiate value from array.
    return sgn


def tresca(s11, s22, s33, s12, s13, s23):
    """Calculate equivalent stress according to Tresca.

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Equivalent Tresca stress. Shape is the same as the components.
    """
    w = eigenval(s11, s22, s33, s12, s13, s23).T
    w_diff = np.zeros(w.shape)
    w_diff[0] = np.fabs(w[0] - w[1])
    w_diff[1] = np.fabs(w[0] - w[2])
    w_diff[2] = np.fabs(w[1] - w[2])
    return np.amax(w_diff, axis=0)


def signed_tresca_trace(s11, s22, s33, s12, s13, s23):
    """Calculate equivalent stress according to Tresca, signed with the sign
    of the trace (i.e s11 + s22 + s33).

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Signed Tresca equivalent stress. Shape is the same as the components.
    """
    return _sign_trace(s11, s22, s33) * tresca(s11, s22, s33, s12, s13, s23)


def signed_tresca_abs_max_principal(s11, s22, s33, s12, s13, s23):
    """Calculate equivalent stress according to Tresca, signed with the sign
    of the absolute maximum principal stress.

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Signed Tresca equivalent stress. Shape is the same as the components.
    """
    return _sign_abs_max_principal(s11, s22, s33, s12, s13, s23) * tresca(s11, s22, s33, s12, s13, s23)


def abs_max_principal(s11, s22, s33, s12, s13, s23):
    """Calculate absolute maximum principal stress (maximum of absolute
    eigenvalues with corresponding sign).

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Absolute maximum principal stress. Shape is the same as the components.
    """
    w = eigenval(s11, s22, s33, s12, s13, s23).T
    w_max = np.amax(w, axis=0)
    w_min = np.amin(w, axis=0)
    sign = _sign_abs_max_principal(s11, s22, s33, s12, s13, s23)
    positive_sign_bool = np.array(sign >= 0)
    return w_max * positive_sign_bool + w_min * np.invert(positive_sign_bool)


def principals(s11, s22, s33, s12, s13, s23):
    """Calculate all principal stress components (eigenvalues).

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        All principal stresses. Shape `(..., 3)`.
    """
    return eigenval(s11, s22, s33, s12, s13, s23)


def max_principal(s11, s22, s33, s12, s13, s23):
    """Calculate maximum principal stress (maximum of eigenvalues).

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Maximum principal stress. Shape is the same as the components.
    """
    w = eigenval(s11, s22, s33, s12, s13, s23).T
    return np.amax(w, axis=0)


def min_principal(s11, s22, s33, s12, s13, s23):
    """Calculate minimum principal stress (minimum of eigenvalues).

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Minimum principal stress. Shape is the same as the components.
    """
    w = eigenval(s11, s22, s33, s12, s13, s23).T
    return np.amin(w, axis=0)


def mises(s11, s22, s33, s12, s13, s23):
    """Calculate equivalent stress according to von Mises.

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Von Mises equivalent stress. Shape is the same as the components.

    Raises
    ------
    AssertionError
        Components' shape is not consistent.
    """
    s11 = np.array(s11)
    s22 = np.array(s22)
    s33 = np.array(s33)
    s12 = np.array(s12)
    s13 = np.array(s13)
    s23 = np.array(s23)

    assert (s11.shape == s22.shape and
            s11.shape == s33.shape and
            s11.shape == s12.shape and
            s11.shape == s13.shape and
            s11.shape == s23.shape), "Components' shape is not consistent."

    mises_stress = np.sqrt(s11 ** 2 + s22 ** 2 + s33 ** 2
                           - s11 * s22 - s11 * s33 - s22 * s33
                           + 3 * (s12 ** 2 + s13 ** 2 + s23 ** 2))
    return mises_stress


def signed_mises_trace(s11, s22, s33, s12, s13, s23):
    """Calculate equivalent stress according to von Mises, signed with the sign
    of the trace (i.e s11 + s22 + s33).

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Signed von Mises equivalent stress. Shape is the same as the components.
    """
    return _sign_trace(s11, s22, s33) * mises(s11, s22, s33, s12, s13, s23)


def signed_mises_abs_max_principal(s11, s22, s33, s12, s13, s23):
    """Calculate equivalent stress according to von Mises, signed with the sign
    of the absolute maximum principal stress.

    Parameters
    ----------
    s11: array_like
        Component 11 of 3D tensor.
    s22: array_like
        Component 22 of 3D tensor.
    s33: array_like
        Component 33 of 3D tensor.
    s12: array_like
        Component 12 of 3D tensor.
    s13: array_like
        Component 13 of 3D tensor.
    s23: array_like
        Component 23 of 3D tensor.

    Returns
    -------
    numpy.ndarray:
        Signed von Mises equivalent stress. Shape is the same as the components.
    """
    return _sign_abs_max_principal(s11, s22, s33, s12, s13, s23) * mises(s11, s22, s33, s12, s13, s23)


@pd.api.extensions.register_dataframe_accessor("equistress")
class StressTensorEquistress(stresssignal.StressTensorVoigt):
    def tresca(self):
        return pd.Series(tresca(s11=self._obj['S11'].to_numpy(),
                                s22=self._obj['S22'].to_numpy(),
                                s33=self._obj['S33'].to_numpy(),
                                s12=self._obj['S12'].to_numpy(),
                                s13=self._obj['S13'].to_numpy(),
                                s23=self._obj['S23'].to_numpy()),
                         name='tresca', index=self._obj.index)

    def signed_tresca_trace(self):
        return pd.Series(signed_tresca_trace(s11=self._obj['S11'].to_numpy(),
                                             s22=self._obj['S22'].to_numpy(),
                                             s33=self._obj['S33'].to_numpy(),
                                             s12=self._obj['S12'].to_numpy(),
                                             s13=self._obj['S13'].to_numpy(),
                                             s23=self._obj['S23'].to_numpy()),
                         name='signed_tresca_trace', index=self._obj.index)

    def signed_tresca_abs_max_principal(self):
        return pd.Series(signed_tresca_abs_max_principal(s11=self._obj['S11'].to_numpy(),
                                                         s22=self._obj['S22'].to_numpy(),
                                                         s33=self._obj['S33'].to_numpy(),
                                                         s12=self._obj['S12'].to_numpy(),
                                                         s13=self._obj['S13'].to_numpy(),
                                                         s23=self._obj['S23'].to_numpy()),
                         name='signed_tresca_abs_max_principal', index=self._obj.index)

    def principals(self):
        all_princ = eigenval(s11=self._obj['S11'].to_numpy(),   # ascending order (numpy.eigvalsh)
                             s22=self._obj['S22'].to_numpy(),
                             s33=self._obj['S33'].to_numpy(),
                             s12=self._obj['S12'].to_numpy(),
                             s13=self._obj['S13'].to_numpy(),
                             s23=self._obj['S23'].to_numpy())
        return pd.DataFrame({'min_principal': all_princ[...,0],
                             'med_principal': all_princ[...,1],
                             'max_principal': all_princ[...,2]},
                             index=self._obj.index)

    def abs_max_principal(self):
        return pd.Series(abs_max_principal(s11=self._obj['S11'].to_numpy(),
                                           s22=self._obj['S22'].to_numpy(),
                                           s33=self._obj['S33'].to_numpy(),
                                           s12=self._obj['S12'].to_numpy(),
                                           s13=self._obj['S13'].to_numpy(),
                                           s23=self._obj['S23'].to_numpy()),
                         name='abs_max_principal', index=self._obj.index)

    def max_principal(self):
        return pd.Series(max_principal(s11=self._obj['S11'].to_numpy(),
                                       s22=self._obj['S22'].to_numpy(),
                                       s33=self._obj['S33'].to_numpy(),
                                       s12=self._obj['S12'].to_numpy(),
                                       s13=self._obj['S13'].to_numpy(),
                                       s23=self._obj['S23'].to_numpy()),
                         name='max_principal', index=self._obj.index)

    def min_principal(self):
        return pd.Series(min_principal(s11=self._obj['S11'].to_numpy(),
                                       s22=self._obj['S22'].to_numpy(),
                                       s33=self._obj['S33'].to_numpy(),
                                       s12=self._obj['S12'].to_numpy(),
                                       s13=self._obj['S13'].to_numpy(),
                                       s23=self._obj['S23'].to_numpy()),
                         name='min_principal', index=self._obj.index)

    def mises(self):
        return pd.Series(mises(s11=self._obj['S11'].to_numpy(),
                               s22=self._obj['S22'].to_numpy(),
                               s33=self._obj['S33'].to_numpy(),
                               s12=self._obj['S12'].to_numpy(),
                               s13=self._obj['S13'].to_numpy(),
                               s23=self._obj['S23'].to_numpy()),
                         name='mises', index=self._obj.index)

    def signed_mises_trace(self):
        return pd.Series(signed_mises_trace(s11=self._obj['S11'].to_numpy(),
                                            s22=self._obj['S22'].to_numpy(),
                                            s33=self._obj['S33'].to_numpy(),
                                            s12=self._obj['S12'].to_numpy(),
                                            s13=self._obj['S13'].to_numpy(),
                                            s23=self._obj['S23'].to_numpy()),
                         name='signed_mises_trace', index=self._obj.index)

    def signed_mises_abs_max_principal(self):
        return pd.Series(signed_mises_abs_max_principal(s11=self._obj['S11'].to_numpy(),
                                                        s22=self._obj['S22'].to_numpy(),
                                                        s33=self._obj['S33'].to_numpy(),
                                                        s12=self._obj['S12'].to_numpy(),
                                                        s13=self._obj['S13'].to_numpy(),
                                                        s23=self._obj['S23'].to_numpy()),
                         name='signed_mises_abs_max_principal', index=self._obj.index)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pytest
import pandas as pd
import numpy as np

from pylife.stress.rainflow.general import AbstractDetector
from pylife.stress.timesignal import TimeSignalGenerator

import pylife.stress.rainflow as RF


class DummyDetector(AbstractDetector):
    def process(self, samples):
        return self


def test_rainflow_partial_get_turns_general():
    tsgen = TimeSignalGenerator(10, {'number': 50,
                                     'amplitude_median': 1.0, 'amplitude_std_dev': 0.5,
                                     'frequency_median': 4, 'frequency_std_dev': 3,
                                     'offset_median': 0, 'offset_std_dev': 0.4},
                                None, None)

    signal_tot = tsgen.query(10000)
    _, turns_tot = DummyDetector(recorder=None)._new_turns(signal_tot)
    rfc_partial = DummyDetector(recorder=None)
    turns_partial = np.concatenate((
        rfc_partial._new_turns(signal_tot[:3424])[1],
        rfc_partial._new_turns(signal_tot[3424:])[1]))

    np.testing.assert_array_equal(turns_tot, turns_partial)


def test_rainflow_partial_signals_get_turns_splitturn():
    tsgen = TimeSignalGenerator(10, {'number': 50,
                                     'amplitude_median': 1.0, 'amplitude_std_dev': 0.5,
                                     'frequency_median': 4, 'frequency_std_dev': 3,
                                     'offset_median': 0, 'offset_std_dev': 0.4},
                                None, None)

    signal_tot = tsgen.query(10000)
    _, turns_tot = DummyDetector(recorder=None)._new_turns(signal_tot)
    turn_points, _ = RF.find_turns(signal_tot)
    turn_points = np.insert(turn_points, 0, 0)
    turn_num = turn_points.shape[0]
    split_points = [int(np.ceil(turn_num*x)) for x in [0.0, 0.137, 0.23, 0.42, 1.0]]
    rfc_partial = DummyDetector(recorder=None)
    turns_partial = np.empty(0)
    for i in range(len(split_points)-1):
        lower = turn_points[split_points[i]]
        upper = 10000 if split_points[i+1] == turn_points.shape[0] else turn_points[split_points[i+1]]
        turns_partial = np.concatenate((turns_partial, rfc_partial._new_turns(signal_tot[lower:upper])[1]))

    print(turns_tot[-5:])
    print(turns_partial[-5:])
    np.testing.assert_array_equal(turns_tot, turns_partial)


def test_rainflow_partial_get_turns_no_turns():
    samples = np.array([0., 1.])
    index, values = RF.find_turns(samples)
    assert len(index) == 0
    assert len(values) == 0


def test_rainflow_partial_get_turns_consecutive_duplicates():
    samples = np.array([1., 1., 0.5, 0.5, 1., 1., 1., -1., -1., 0.5, 1.])
    index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(values, np.array([0.5, 1., -1.]))
    np.testing.assert_array_equal(index, np.array([2, 4, 7]))


def test_rainflow_duplicates_no_peak_up():
    samples = np.array([1., 2., 2., 3.])
    index, values = RF.find_turns(samples)
    assert len(index) == 0
    assert len(values) == 0


def test_rainflow_duplicates_no_peak_down():
    samples = np.array([3., 2., 2., 1.])
    index, values = RF.find_turns(samples)
    assert len(index) == 0
    assert len(values) == 0


def test_rainflow_find_turns_shifted_index():
    samples = np.array([32., 32., 32.1, 32.9, 33., 33., 33., 33., 33., 32.5, 32., 32., 32.7, 37.2, 40., 35.2, 33.])
    expected_index = [4, 10, 14]
    expected_values = [33., 32., 40.]
    index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)


def test_rainflow_find_turns_shifted_index_four_leading_dups():
    samples = np.array([32., 32., 32., 32., 32.1, 32.9, 33., 33., 33., 33., 33., 32.5, 32., 32., 32.7, 37.2, 40., 35.2, 33.])
    expected_index = [6, 12, 16]
    expected_values = [33., 32., 40.]
    index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)


def test_rainflow_find_turns_shifted_index_four_trailing_dups():
    samples = np.array([32., 32.1, 32.9, 33., 33., 33., 33., 33., 32.5, 32., 32., 32.7, 37.2, 40., 35.2, 33., 33., 33., 33., 33.])
    expected_index = [3, 9, 13]
    expected_values = [33., 32., 40.]
    index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)


def test_find_turns_leading_dups():
    samples = np.array([1., 1., 1., 2., 1.])
    expected_index = [3]
    expected_values = [2.]

    index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)


def test_find_turns_trailing_dups():
    samples = np.array([1., 2., 1., 1., 1.])
    expected_index = [1]
    expected_values = [2.]

    index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)


def test_find_turns_leading_and_trailing_dups():
    samples = np.array([0., 0., 1., 2., 2., 1., 0., 0., 1., 2., 2.])
    expected_index = [3, 6]
    expected_values = [2., 0.]

    index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)


def test_find_turns_leading_and_trailing_dups_no_turns():
    samples = np.array([0.0, 0.0, 0.5, 1.0, 1.0])
    expected_index = []
    expected_values = []

    index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)


def test_find_turns_flat_signal():
    samples = np.array([0.0, 0.0, 0.0, 0.0])
    expected_index = []
    expected_values = []

    index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)


@pytest.mark.parametrize('null', [None, np.nan, pd.NaT, float("nan"), pd.NA])
def test_find_turns_nan_next_to_turns_nan_after_turn(null):
    samples = np.array([0.0, 1.0, null, 0.0, null, 1.0])
    expected_index = [1, 3]
    expected_values = [1.0, 0.0]

    with pytest.warns(UserWarning, match="At least one NaN like value has been dropped from the input signal."):
        index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)


@pytest.mark.parametrize('null', [None, np.nan, pd.NaT, float("nan"), pd.NA])
def test_find_turns_nan_next_to_turns_nan_before_turns(null):
    samples = np.array([0.0, null, 1.0, null, 0.0, 1.0])
    expected_index = [2, 4]
    expected_values = [1.0, 0.0]

    with pytest.warns(UserWarning, match="At least one NaN like value has been dropped from the input signal."):
        index, values = RF.find_turns(samples)
    np.testing.assert_array_equal(index, expected_index)
    np.testing.assert_array_equal(values, expected_values)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



import pytest
import numpy as np
import pandas as pd

import pylife.strength.fatigue
import pylife.stress


def load_index():
    nCode_Xbinsize = 62.375
    nCode_XMax = 468.8125
    nCode_XMin = 32.1875

    return pd.interval_range(start=nCode_XMin-nCode_Xbinsize/2,
                             end=nCode_XMax+nCode_Xbinsize/2, periods=8, name="range")


expected_elementary_1 = pd.Series([
    2.057849E-06,
    3.039750E-05,
    1.507985E-05,
    6.434856E-06,
    6.296737E-06,
    1.751881E-06,
    0.000000E+00,
    0.000000E+00
], name='damage', index=load_index())
expected_elementary_2 = pd.Series([
    1.868333E-06,
    3.894329E-05,
    4.829155E-05,
    1.543643E-05,
    1.357734E-05,
    1.379607E-05,
    1.023415E-05,
    7.548524E-07
], name='damage', index=load_index())
expected_elementary_3 = pd.Series([
    3.067487E-06,
    1.999014E-08,
    0.000000E+00,
    0.000000E+00,
    0.000000E+00,
    0.000000E+00,
    0.000000E+00,
    0.000000E+00
], name='damage', index=load_index())

expected_haibach_1 = pd.Series([
    2.293973E-08,
    7.414563E-05,
    4.631857E-04,
    4.779392E-04,
    6.006944E-04,
    2.041326E-04,
    0.000000E+00,
    0.000000E+00
], name='damage', index=load_index())
expected_haibach_2 = pd.Series([
    2.083222E-08,
    9.499395E-05,
    1.483300E-03,
    1.146517E-03,
    1.295247E-03,
    1.607545E-03,
    1.408691E-03,
    1.198482E-04
], name='damage', index=load_index())
expected_haibach_3 = pd.Series([
    3.419927E-08,
    4.876799E-08,
    0.000000E+00,
    0.000000E+00,
    0.000000E+00,
    0.000000E+00,
    0.000000E+00,
    0.000000E+00
], name='damage', index=load_index())

expected_original_1 = pd.Series([
    0.00000000E+00,
    0.00000000E+00,
    2.08681840E-05,
    1.73881920E-05,
    2.80698130E-05,
    1.16511320E-05,
    0.00000000E+00,
    0.00000000E+00,
], name='damage', index=load_index())
expected_original_2 = pd.Series([
    0.00000000E+00,
    0.00000000E+00,
    6.68280320E-05,
    4.17121220E-05,
    6.05255350E-05,
    9.17526660E-05,
    9.49790820E-05,
    9.32071420E-06
], name='damage', index=load_index())
expected_original_3 = pd.Series([
    0.00000000E+00,
    0.00000000E+00,
    0.00000000E+00,
    0.00000000E+00,
    0.00000000E+00,
    0.00000000E+00,
    0.00000000E+00,
    0.00000000E+00
], name='damage', index=load_index())


material = pd.DataFrame(
    index=['k_1', 'ND', 'SD'],
    columns=['elementary', 'haibach', 'original'],
    data=[[4., 5., 6.], [4e7, 1e6, 1e8], [100., 90., 75.]]
)

load_hist_1 = pd.Series([
    1.227E5,
    2.433E4,
    1591.0,
    178.0,
    64.0,
    8.0,
    0.0,
    0.0,
], name='cycles', index=load_index())

load_hist_2 = pd.Series([
    1.114E5,
    3.117E4,
    5095.0,
    427.0,
    138.0,
    63.0,
    24.0,
    1.0,
], name='cycles', index=load_index())

load_hist_3 = pd.Series([
    1.829E5,
    16.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0
], name='cycles', index=load_index())


@pytest.mark.parametrize('load_hist, expected', [
    (load_hist_1, expected_elementary_1),
    (load_hist_2, expected_elementary_2),
    (load_hist_3, expected_elementary_3),
])
def test_damage_fatigue_elementary(load_hist, expected):
    fatigue = material['elementary'].fatigue.miner_elementary().damage(load_hist.load_collective)
    pd.testing.assert_series_equal(fatigue, expected, rtol=1e-3)


@pytest.mark.parametrize('load_hist, expected', [
    (load_hist_1, expected_haibach_1),
    (load_hist_2, expected_haibach_2),
    (load_hist_3, expected_haibach_3),
])
def test_damage_fatigue_haibach(load_hist, expected):
    fatigue = material['haibach'].fatigue.miner_haibach().damage(load_hist.load_collective)
    pd.testing.assert_series_equal(fatigue, expected, rtol=1e-3)


@pytest.mark.parametrize('load_hist, expected', [
    (load_hist_1, expected_original_1),
    (load_hist_2, expected_original_2),
    (load_hist_3, expected_original_3),
])
def test_damage_fatigue_original(load_hist, expected):
    fatigue = material['original'].fatigue.damage(load_hist.load_collective)
    pd.testing.assert_series_equal(fatigue, expected, rtol=1e-3)


@pytest.mark.parametrize('TS, allowed_pf, expected', [
    (2., 0.5, 5.0),
    (1.0000001, 0.1, 5.0),
    (1.25, 1e-6, 3.3055576111)
])
def test_security_load_single(TS, allowed_pf, expected):
    wc = pd.Series({
        'SD': 500.,
        'k_1': 6.0,
        'TS': TS,
        'ND': 1e6
    })
    load_collective = pd.DataFrame({'mean': [0.0], 'amplitude': [100.0], 'cycles': 1e7})
    security_factor = wc.fatigue.security_load(load_collective, allowed_pf)
    np.testing.assert_allclose(security_factor, expected)


def test_security_load_multiple():
    idx = pd.Index([1, 2, 3], name='element_id')
    wc = pd.DataFrame({
        'SD': [500., 500., 300.],
        'k_1': [6., 6., 6.],
        'TS': [2., 1.0000001, 1.25],
        'ND': [1e6, 1e6, 1e6]
    }, index=idx)
    load_signal = pd.DataFrame({
        'mean': [0., 0., 0.],
        'amplitude': [250., 125.,  100.],
        'cycles': [1e7, 1e7, 1e7]
    }, index=idx)

    expected = pd.Series([2.0, 4.0, 3.0], index=idx, name='security_factor')
    result = wc.fatigue.security_load(load_signal, 0.5)

    pd.testing.assert_series_equal(result, expected)



@pytest.mark.parametrize('TN, allowed_pf, cycles, expected', [
    (4., 0.5, 1e5, 10.0),
    (4., 0.5, 1e4, 100.0),
    (4., 0.1, 1e4, 50.0),
    (4., 1e-6, 1e4, 7.65),
    (4., 1e-6, 1e3, 76.5),
])
def test_security_cycles_single(TN, allowed_pf, cycles, expected):
    wc = pd.Series({
        'SD': 500.,
        'k_1': 0.5,
        'TN': TN,
        'ND': 1e6
    })
    load_collective = pd.DataFrame({'mean': [0.0], 'amplitude': [500.0], 'cycles': [cycles]})
    security_factor = wc.fatigue.security_cycles(load_collective, allowed_pf)
    np.testing.assert_allclose(security_factor, expected, rtol=1e-3)


def test_security_cycles_multiple():
    idx = pd.Index([1, 2, 3], name='element_id')
    wc = pd.DataFrame({
        'SD': [500., 500., 500.],
        'k_1': [.5, .5, .5],
        'TN': [4., 4., 4.],
        'ND': [1e6, 1e6, 1e6],
    }, index=idx)

    load_collective = pd.DataFrame({
        'amplitude': [500., 500., 500.],
        'cycles': [1e6, 1e5, 1e4],
    }, index=idx)

    expected = pd.Series([1., 10., 100.], index=idx, name='security_factor')
    result = wc.fatigue.security_cycles(load_collective, 0.5)

    pd.testing.assert_series_equal(result, expected)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import unittest
import numpy as np
import pandas as pd

import pylife.stress.rainflow as RF
import pylife.stress.rainflow.recorders as RFR


def process_signal(signal):
    fr = RFR.FullRecorder()
    dtor = RF.FKMDetector(recorder=fr).process(signal)

    return fr, dtor


class TestFKMMemory1Inner(unittest.TestCase):

    def setUp(self):
        signal = np.array([0., 100., 0., 80., 20., 60., 40., 100., 0., 80., 20., 60., 40., 45.])
        self._fr, self._dtor = process_signal(signal)

    def test_values(self):
        np.testing.assert_array_equal(self._fr.values_from, np.array([60., 80., 100.]))
        np.testing.assert_array_equal(self._fr.values_to, np.array([40., 20., 0.]))

    def test_residuals(self):
        np.testing.assert_array_equal(self._dtor.residuals, np.array([100., 0., 80., 20., 60., 40.]))


class TestFKMMemory1_2_3(unittest.TestCase):

    def setUp(self):
        signal = np.array([0.,
                           1., -1., 1., -2., -1., -2., 2., 0., 2., -2.,
                           1., -1., 1., -2., -1., -2., 2., 0., 2., -2.,
                           -1.8])
        self._fr, self._dtor = process_signal(signal)

    def test_values(self):
        np.testing.assert_array_equal(self._fr.values_from, np.array([1., -2., 2., -2.,  1., -2., -2., 2., -2.]))
        np.testing.assert_array_equal(self._fr.values_to, np.array([-1., -1., 0.,  2., -1.,  1., -1., 0.,  2.]))

    def test_residuals(self):
        np.testing.assert_array_equal(self._dtor.residuals, np.array([1., -2.]))


def test_series_signal_float_index():
    signal = pd.Series([0., 1., 0.], index=[1.0, 2.0, 3.0])
    process_signal(signal)



# Copyright (c) 2019-2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from .elementary import Elementary

import numpy as np

import pylife.utils.functions as functions
from pylife.utils.probability_data import ProbabilityFit


class Probit(Elementary):
    """Wöhler analysis according to the Probit method.

    For each load level in the infinite regime a failure probability is
    estimated.  To these failure probability a log norm distribution is fitted,
    whose parameters are then used to calculate the Wöhler curve parameters.
    """

    def _specific_analysis(self, wc):
        wc_init = wc.copy(deep=True)
        self._inf_groups = self._fd.infinite_zone.groupby('load')
        if len(self._inf_groups) < 2:
            warnings.warn(UserWarning("Probit needs at least two – preferably mixed – load levels in the infinite zone. Falling back to Elementary."))
            return wc

        wc['TS'], wc['SD'], wc['ND'] = self.__probit_analysis(wc_init)
        return wc

    def __probit_rossow_estimation(self):
        frac_num = self._inf_groups.fracture.sum().astype(int).to_numpy()
        tot_num = self._inf_groups.fracture.count().to_numpy()

        fprobs = np.empty_like(frac_num, dtype=np.float64)
        no_fractures = frac_num == 0
        w = np.where(no_fractures)
        fprobs[w] = 1. - 0.5**(1./tot_num[w])

        all_fractures = frac_num == tot_num
        w = np.where(all_fractures)
        fprobs[w] = 0.5**(1./tot_num[w])

        some_fractures = np.logical_and(np.logical_not(no_fractures), np.logical_not(all_fractures))
        w = np.where(some_fractures)
        fprobs[w] = (3*frac_num[w] - 1) / (3*tot_num[w] + 1)

        return fprobs, self._inf_groups.load.mean()

    def __probit_analysis(self, wc_init):
        if self._fd.num_runouts == 0:
            return 1., 0., self._transition_cycles(0.0)

        fprobs, load = self.__probit_rossow_estimation()

        self._probability_fit = ProbabilityFit(fprobs, load)

        TS = functions.std_to_scattering_range(1./self._probability_fit.slope)
        SD = 10**(-self._probability_fit.intercept/self._probability_fit.slope)

        ND = np.nan if np.isnan(wc_init['ND']) else self._transition_cycles(SD)

        return TS, SD, ND



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'Alexander Maier'
__maintainer__ = __author__

import numpy as np


class _Hookeslawcore:
    '''Parent class for the multidimensional Hooke's Law implementation. Defines the properties and checks for correct inputs'''

    def __init__(self, E, nu):
        '''Instantiate a multidimensional Hooke's Law implementation

        Parameters
        ----------

        E : float
            Young's modulus

        nu : float
            Poisson's ratio. Must be between -1 and 1./2.
        '''
        self._validateinit(nu)
        self._E = E
        self._nu = nu
        self._G = E / (2. * (1 + nu))
        self._K = E / (3. * (1 - 2 * nu))

    def _validateinit(self, nu):
        '''Validates the input of the Poisson\'s ratio
        '''
        if nu < - 1 or nu > 1./2:
            raise ValueError('Poisson\'s ratio nu is %.2f but must be -1 <= nu <= 1./2.' % nu)

    def _as_consistant_arrays(self, *args):
        '''Transforms the inputs into numpy arrays and checks the shape of the given inputs. If the shapes are note equal, a ValueError is raised
        '''
        transformed = tuple(np.asarray(arg) for arg in args)
        shape0 = transformed[0].shape
        shape = [shape0 == arg.shape for arg in transformed]
        if not all(shape):
            raise ValueError('Components\' shape is not consistent.')

        return transformed

    @property
    def E(self):
        '''Get Young's modulus'''
        return self._E

    @property
    def nu(self):
        '''Get Poisson's ratio'''
        return self._nu

    @property
    def G(self):
        '''Get the sheer modulus'''
        return self._G

    @property
    def K(self):
        '''Get the bulk modulus'''
        return self._K


class HookesLaw1d:
    '''Implementation of the one dimensional Hooke's Law

    Parameters
    ----------

    E : float
        Young's modulus
    '''

    def __init__(self, E):
        '''
        Instantiate a one dimensional Hooke's Law implementation with a given Young's modulus

        Parameters
        ----------

        E : float
            Young's modulus 
        '''
        self._E = E

    @property
    def E(self):
        '''Get Young's modulus'''
        return self._E

    def stress(self, strain):
        '''Get the stress for a given elastic strain

        Parameters
        ----------
        strain : array-like float
            The elastic strain

        Returns
        -------
        strain : array-like float
            The resulting stress
        '''
        return np.asarray(strain) * self._E

    def strain(self, stress):
        '''Get the elastic strain for a given stress

        Parameters
        ----------
        stress : array-like float
            The stress

        Returns
        -------
        strain : array-like float
            The resulting elastic strain
        '''
        return np.asarray(stress) / self._E


class HookesLaw2dPlaneStress(_Hookeslawcore):
    '''Implementation of the Hooke's Law under plane stress conditions.

    Parameters
    ----------

    E : float
        Young's modulus

    nu : float
        Poisson's ratio. Must be between -1 and 1./2.

    Notes
    -----

    A cartesian coordinate system is assumed. The stress components in 3 direction are assumed to be zero, s33 = s13 = s23 = 0.'''

    def __init__(self, E, nu):
        super().__init__(E, nu)
        self._Et = E
        self._nut = self._nu

    def strain(self, s11, s22, s12):
        '''Get the elastic strain components for given stress components

        Parameters
        ----------
        s11 : array-like float
            The normal stress component with basis 1-1
        s22 : array-like float
            The normal stress component with basis 2-2
        s12 : array-like float
            The shear stress component with basis 1-2


        Returns
        -------
        e11 : array-like float
            The resulting elastic normal strain component with basis 1-1
        e22 : array-like float
            The resulting elastic normal strain component with basis 2-2
        e33 : array-like float
            The resulting elastic normal strain component with basis 3-3
        g12 : array-like float
            The resulting elastic engineering shear strain component with basis 1-2, 
            (1. / 2 * g12 is the tensor component)
        '''
        s11, s22, s12 = self._as_consistant_arrays(s11, s22, s12)
        e11 = 1. / self._Et * (s11 - self._nut * s22)
        e22 = 1. / self._Et * (s22 - self._nut * s11)
        e33 = - self._nu / self._E * (s11 + s22)
        g12 = 1. / self._G * s12
        return e11, e22, e33, g12

    def stress(self, e11, e22, g12):
        '''Get the stress components for given elastic strain components

        Parameters
        ----------
        e11 : array-like float
            The elastic normal strain component with basis 1-1
        e22 : array-like float
            The elastic normal strain component with basis 2-2
        g12 : array-like float
            The elastic engineering shear strain component with basis 1-2, 
            (1. / 2 * g12 is the tensor component)

        Returns
        -------
        s11 : array-like float
            The resulting normal stress component with basis 1-1
        s22 : array-like float
            The resulting normal stress component with basis 2-2
        s12 : array-like float
            The resulting shear stress component with basis 1-2
        '''
        e11, e22, g12 = self._as_consistant_arrays(e11, e22, g12)
        factor = self._Et / (1 - np.power(self._nut, 2.))
        s11 = factor * (e11 + self._nut * e22)
        s22 = factor * (e22 + self._nut * e11)
        s12 = self._G * g12
        return s11, s22, s12


class HookesLaw2dPlaneStrain(HookesLaw2dPlaneStress):
    '''Implementation of the Hooke's Law under plane strain conditions. 

    Parameters
    ----------

    E : float
        Young's modulus

    nu : float
        Poisson's ratio. Must be between -1 and 1./2.

    Notes
    -----

    A cartesian coordinate system is assumed. The strain components in 3 direction are assumed to be zero, e33 = g13 = g23 = 0.
    '''

    def __init__(self, E, nu):
        super().__init__(E, nu)
        self._Et = self._E / (1 - np.power(self._nu, 2))
        self._nut = self._nu / (1 - self._nu)

    def strain(self, s11, s22, s12):
        '''Get the elastic strain components for given stress components

        Parameters
        ----------
        s11 : array-like float
            The normal stress component with basis 1-1
        s22 : array-like float
            The normal stress component with basis 2-2
        s12 : array-like float
            The shear stress component with basis 1-2


        Returns
        -------
        e11 : array-like float
            The resulting elastic normal strain component with basis 1-1
        e22 : array-like float
            The resulting elastic normal strain component with basis 2-2
        g12 : array-like float
            The resulting elastic engineering shear strain component with basis 1-2, 
            (1. / 2 * g12 is the tensor component)
        '''
        e11, e22, _, g12 = super().strain(s11, s22, s12)
        return e11, e22, g12

    def stress(self, e11, e22, g12):
        '''Get the stress components for given elastic strain components

        Parameters
        ----------
        e11 : array-like float
            The elastic normal strain component with basis 1-1
        e22 : array-like float
            The elastic normal strain component with basis 2-2
        g12 : array-like float
            The elastic engineering shear strain component with basis 1-2, 
            (1. / 2 * g12 is the tensor component)

        Returns
        -------
        s11 : array-like float
            The resulting normal stress component with basis 1-1
        s22 : array-like float
            The resulting normal stress component with basis 2-2
        s33 : array-like float
            The resulting normal stress component with basis 3-3
        s12 : array-like float
            The resulting shear stress component with basis 1-2
        '''
        s11, s22, s12 = super().stress(e11, e22, g12)
        s33 = self.nu * (s11 + s22)
        return s11, s22, s33, s12


class HookesLaw3d(_Hookeslawcore):
    '''Implementation of the Hooke's Law in three dimensions.

    Parameters
    ----------

    E : float
        Young's modulus

    nu : float
        Poisson's ratio. Must be between -1 and 1./2

    Notes
    -----

    A cartesian coordinate system is assumed.
    '''

    def __init__(self, E, nu):
        super().__init__(E, nu)

    def strain(self, s11, s22, s33, s12, s13, s23):
        '''Get the elastic strain components for given stress components

        Parameters
        ----------
        s11 : array-like float
            The resulting normal stress component with basis 1-1
        s22 : array-like float
            The resulting normal stress component with basis 2-2
        s33 : array-like float
            The resulting normal stress component with basis 3-3
        s12 : array-like float
            The resulting shear stress component with basis 1-2
        s13 : array-like float
            The resulting shear stress component with basis 1-3
        s23 : array-like float
            The resulting shear stress component with basis 2-3

        Returns
        -------
        e11 : array-like float
            The resulting elastic normal strain component with basis 1-1
        e22 : array-like float
            The resulting elastic normal strain component with basis 2-2
        e33 : array-like float
            The resulting elastic normal strain component with basis 3-3
        g12 : array-like float
            The resulting elastic engineering shear strain component with basis 1-2, 
            (1. / 2 * g12 is the tensor component)
        g13 : array-like float
            The resulting elastic engineering shear strain component with basis 1-3, 
            (1. / 2 * g13 is the tensor component)
        g23 : array-like float
            The resulting elastic engineering shear strain component with basis 2-3, 
            (1. / 2 * g23 is the tensor component)
        '''
        s11, s22, s33, s12, s13, s23 = self._as_consistant_arrays(s11, s22, s33, s12, s13, s23)
        e11 = 1 / self._E * (s11 - self._nu * (s22 + s33))
        e22 = 1 / self._E * (s22 - self._nu * (s11 + s33))
        e33 = 1 / self._E * (s33 - self._nu * (s11 + s22))
        g12 = s12 / self._G
        g13 = s13 / self._G
        g23 = s23 / self._G
        return e11, e22, e33, g12, g13, g23

    def stress(self, e11, e22, e33, g12, g13, g23):
        '''Get the stress components for given elastic strain components

        Parameters
        ----------
        e11 : array-like float
            The elastic normal strain component with basis 1-1
        e22 : array-like float
            The elastic normal strain component with basis 2-2
        e33 : array-like float
            The elastic normal strain component with basis 3-3
        g12 : array-like float
            The elastic engineering shear strain component with basis 1-2, 
            (1. / 2 * g12 is the tensor component)
        g13 : array-like float
            The elastic engineering shear strain component with basis 1-3, 
            (1. / 2 * g13 is the tensor component)
        g23 : array-like float
            The elastic engineering shear strain component with basis 2-3, 
            (1. / 2 * g23 is the tensor component)

        Returns
        -------
        s11 : array-like float
            The resulting normal stress component with basis 1-1
        s22 : array-like float
            The resulting normal stress component with basis 2-2
        s33 : array-like float
            The resulting normal stress component with basis 3-3
        s12 : array-like float
            The resulting shear stress component with basis 1-2
        s13 : array-like float
            The resulting shear stress component with basis 1-3
        s23 : array-like float
            The resulting shear stress component with basis 2-3
        '''
        e11, e22, e33, g12, g13, g23 = self._as_consistant_arrays(e11, e22, e33, g12, g13, g23)
        factor1 = self._E / ((1 + self._nu) * (1 - 2 * self._nu))
        factor2 = 1 - self._nu
        s11 = factor1 * (factor2 * e11 + self._nu * (e22 + e33))
        s22 = factor1 * (factor2 * e22 + self._nu * (e11 + e33))
        s33 = factor1 * (factor2 * e33 + self._nu * (e11 + e22))
        s12 = self._G * g12
        s13 = self._G * g13
        s23 = self._G * g23
        return s11, s22, s33, s12, s13, s23



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pandas as pd

from pylife.materiallaws import WoehlerCurve


@pd.api.extensions.register_series_accessor('fatigue')
@pd.api.extensions.register_dataframe_accessor('fatigue')
class Fatigue(WoehlerCurve):
    """Extension for ``WoehlerCurve`` accessor class for fatigue calculations.

    Note
    ----
    This class is accessible by the ``fatigue`` accessor attribute.
    """

    def damage(self, load_collective):
        """Calculate the damage to the material caused by a given load collective.

        Parameters
        ----------
        load_collective : pandas object or object behaving like a load collective
            The given load collective

        Returns
        -------
        damage : :class:`pandas.Series`
            The calculated damage values. The index is the broadcast between
            ``load_collective`` and ``self``.
        """
        cycles = self.cycles(load_collective.amplitude)
        return pd.Series(load_collective.cycles / cycles, name='damage')

    def security_load(self, load_distribution, allowed_failure_probability):
        """Calculate the security factor in load direction for given load distribution.

        Parameters
        ----------
        load_distribution : pandas object or object behaving like a load collective
            The given load distribution

        Returns
        -------
        security_factor : :class:`pandas.Series`
            The calculated security_factors. The index is the broadcast between
            ``load_distribution`` and ``self``.
        """
        allowed_load = self.load(load_distribution.cycles, allowed_failure_probability)
        return pd.Series(allowed_load / load_distribution.amplitude, name='security_factor')

    def security_cycles(self, load_distribution, allowed_failure_probability):
        """Calculate the security factor in cycles direction for given load distribution.

        Parameters
        ----------
        load_distribution : pandas object or object behaving like a load collective
            The given load distribution

        Returns
        -------
        security_factor : :class:`pandas.Series`
            The calculated security_factors. The index is the broadcast between
            ``load_distribution`` and ``self``.
        """
        allowed_cycles = self.cycles(load_distribution.amplitude, allowed_failure_probability)
        return pd.Series(allowed_cycles / load_distribution.cycles, name='security_factor')



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module for Wöhler curve fatigue data analysis

Overview
========

:class:`FatigueData` is a signal accessor class to handle fatigue data from a
Wöhler test.  They can be analyzed by several analyzers according to your choice

* :class:`Elementary` only treats the finite zone of the fatigue data and
  calculates the slope and the scatter in lifetime direction.  It is the base
  class for all other analyzers

* :class:`Probit` calculates parameters not calculated by :class:`Elementary`
  using the Probit method.

* :class:`MaxLikeInf` calculates parameters not calculated by :class:`Elementary`
  using the maximum likelihood method.

* :class:`MaxLikeFull` calculates all parameters using the maximum likelihood
  method.  The result from :class:`Elementary` is used as start values.

"""

__author__ = "Johannes Mueller"
__maintainer__ = __author__

from .fatigue_data import \
    FatigueData, \
    determine_fractures

from .elementary import Elementary
from .probit import Probit
from .maxlike import MaxLikeInf, MaxLikeFull

__all__ = [
    'FatigueData',
    'determine_fractures',
    'Elementary',
    'Probit',
    'MaxLikeInf',
    'MaxLikeFull',
]




import numpy as np
import pylife.stress.timesignal as TS

np.random.seed(23424711)

load_signal = TS.TimeSignalGenerator(
    10,
    {
        'number': 50_000,
        'amplitude_median': 50.0,
        'amplitude_std_dev': 0.5,
        'frequency_median': 4,
        'frequency_std_dev': 3,
        'offset_median': 0,
        'offset_std_dev': 0.4,
    },
    None,
    None,
).query(1_000_000)

np.savetxt('load_signal.csv', load_signal)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = ["Johannes Mueller", 'Alexander Maier']
__maintainer__ = __author__

import numpy as np
from scipy import optimize


class RambergOsgood:
    '''Simple implementation of the Ramberg-Osgood relation

    Parameters
    ----------

    E : float
        Young's Modulus
    K : float
        The strength coefficient
    n : float
        The strain hardening coefficient

    Notes
    -----
    The equation implemented is the one that `Wikipedia
    <https://en.wikipedia.org/wiki/Ramberg%E2%80%93Osgood_relationship#Alternative_Formulations>`__
    refers to as "Alternative Formulation". The parameters `n` and `k` in this
    are formulation are the Hollomon parameters.

    '''

    def __init__(self, E, K, n):
        self._E = E
        self._K = K
        self._n = n

    @property
    def E(self):
        '''Get Young's Modulus'''
        return self._E

    @property
    def K(self):
        '''Get the strength coefficient'''
        return self._K

    @property
    def n(self):
        '''Get the strain hardening coefficient'''
        return self._n

    def strain(self, stress):
        '''Calculate the elastic plastic strain for a given stress

        Parameters
        ----------
        stress : array-like float
            The stress

        Returns
        -------
        strain : array-like float
            The resulting strain

        '''
        stress = np.asarray(stress)
        return self.elastic_strain(stress) + self.plastic_strain(stress)

    def elastic_strain(self, stress):
        '''Calculate the elastic strain for a given stress

        Parameters
        ----------
        stress : array-like float
            The stress

        Returns
        -------
        strain : array-like float
            The resulting elastic strain
        '''
        return stress/self._E

    def plastic_strain(self, stress):
        '''Calculate the plastic strain for a given stress

        Parameters
        ----------
        stress : array-like float
            The stress

        Returns
        -------
        strain : array-like float
            The resulting plastic strain
        '''
        absstress, signstress = self._get_abs_sign(stress)
        return signstress * np.power(absstress/self._K, 1./self._n)

    def _get_abs_sign(self, x):
        '''Calculate the absolute value and the sign for a given input

        Parameters
        ----------
        x : array-like float
            The input

        Returns
        -------
        abs_x : array-like float
            The resulting absolute value
        sign_x : array-like float
            The resulting sign of the input
        '''
        abs_x = np.fabs(x)
        sign_x = np.sign(x)
        return abs_x, sign_x

    def stress(self, strain, rtol=1e-5, tol=1e-6):
        '''Calculate the stress for a given strain

        Parameters
        ----------
        strain : array-like float
            The strain

        Returns
        -------
        stress : array-like float
            The resulting stress
        '''

        def residuum(stress):
            return self.strain(stress) - abs_strain

        def dresiduum(stress):
            return self.tangential_compliance(stress)

        strain = np.asarray(strain)
        abs_strain, sign_strain = self._get_abs_sign(strain)
        stress0 = self._E * abs_strain
        abs_stress = optimize.newton(func=residuum, x0=stress0, fprime=dresiduum, rtol=rtol, tol=tol)
        return abs_stress * sign_strain

    def tangential_compliance(self, stress):
        '''Calculate the derivative of the strain with respect to the stress for a given stress

        Parameters
        ----------
        stress : array-like float
            The stress

        Returns
        -------
        dstrain : array-like float
            The resulting derivative
        '''
        stress = np.abs(stress)
        return 1./self._E + 1./(self._n*self._K) * np.power(stress/self._K, 1./self._n - 1)

    def tangential_modulus(self, stress):
        '''Calculate the derivative of the stress with respect to the strain for a given stress

        Parameters
        ----------
        stress : array-like float
            The stress

        Returns
        -------
        dstress : array-like float
            The resulting derivative
        '''
        return 1. / self.tangential_compliance(stress)

    def delta_strain(self, delta_stress):
        '''Calculate the cyclic Masing strain span for a given stress span

        Parameters
        ----------
        delta_stress : array-like float
            The stress span

        Returns
        -------
        delta_strain : array-like float
            The corresponding strain span

        Notes
        -----
        A Masing like behavior is assumed for the material as described in
        `Kerbgrundkonzept <https://de.wikipedia.org/wiki/Kerbgrundkonzept#Masing-Verhalten_und_Werkstoffged%C3%A4chtnis>`__.
        '''
        return 2*self.strain(stress=delta_stress/2.)

    def delta_stress(self, delta_strain):
        '''Calculate the cyclic Masing stress span for a given strain span

        Parameters
        ----------
        delta_strain : array-like float
            The strain span

        Returns
        -------
        delta_stress : array-like float
            The corresponding stress span

        Notes
        -----
        A Masing like behavior is assumed for the material as described in
        `Kerbgrundkonzept <https://de.wikipedia.org/wiki/Kerbgrundkonzept#Masing-Verhalten_und_Werkstoffged%C3%A4chtnis>`__.
        '''
        return 2*self.stress(strain=delta_strain/2.)

    def lower_hysteresis(self, stress, max_stress):
        '''Calculate the lower (relaxation to compression) hysteresis starting from a given maximum stress

        Parameters
        ----------
        stress : array-like float
            The stress (must be below the maximum stress)
        max_stress : float
            The maximum stress of the hysteresis look

        Returns
        -------
        lower_hysteresis : array-like float
            The lower hysteresis branch from `max_stress` all the way to `stress`

        Raises
        ------
        ValueError
            if stress > max_stress
        '''
        stress = np.asarray(stress)
        if (stress > max_stress).any():
            raise ValueError("Value for 'stress' must not be higher than 'max_stress'.")
        return self.strain(max_stress) - self.delta_strain(max_stress-stress)



import re
import pytest
import numpy as np

from pylife.materiallaws import RambergOsgood


@pytest.fixture
def ramberg_osgood_monotone():
    return RambergOsgood(E=2., K=6., n=0.5)


def test_rambgood_init(ramberg_osgood_monotone):
    assert ramberg_osgood_monotone._E == 2.
    assert ramberg_osgood_monotone._K == 6.
    assert ramberg_osgood_monotone._n == 0.5


def test_rambgood_properties(ramberg_osgood_monotone):
    assert ramberg_osgood_monotone.E == ramberg_osgood_monotone._E
    assert ramberg_osgood_monotone.K == ramberg_osgood_monotone._K
    assert ramberg_osgood_monotone.n == ramberg_osgood_monotone._n


parametrization_data_monotone = np.array([
    [0.0, 0.0],
    [1.0, 19./36.],
    [2.0, 10./9.],
    [3.0, 7./4.],
    [-1.0, -19./36.],
    [-2.0, -10./9.],
    [-3.0, -7./4.]
])


@pytest.mark.parametrize('stress, expected', map(tuple, parametrization_data_monotone))
def test_rambgood_strain_scalar(ramberg_osgood_monotone, stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood_monotone.strain(stress), expected, significant=5)


@pytest.mark.parametrize('expected, strain', map(tuple, parametrization_data_monotone))
def test_rambgood_stress_scalar(ramberg_osgood_monotone, expected, strain):
    np.testing.assert_approx_equal(ramberg_osgood_monotone.stress(strain), expected, significant=5)


@pytest.mark.parametrize('stress, expected', [
    (parametrization_data_monotone[:, 0], parametrization_data_monotone[:, 1])
])
def test_rambgood_strain_array(ramberg_osgood_monotone, stress, expected):
    np.testing.assert_allclose(ramberg_osgood_monotone.strain(stress), expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [
    (parametrization_data_monotone[:, 0], parametrization_data_monotone[:, 1])
])
def test_rambgood_stress_array(ramberg_osgood_monotone, expected, strain):
    np.testing.assert_allclose(ramberg_osgood_monotone.stress(strain), expected, rtol=1e-5)


parametrization_data_monotone_plastic = np.array([
    [0.0, 0.0],
    [1.0, 1./36.],
    [2.0, 1./9.],
    [3.0, 1./4.],
    [-1.0, -1./36.],
    [-2.0, -1./9.],
    [-3.0, -1./4.]
])


@pytest.mark.parametrize('stress, expected', [
    (parametrization_data_monotone_plastic[:, 0], parametrization_data_monotone_plastic[:, 1])
])
def test_rambgood_plastic_strain_scalar(ramberg_osgood_monotone, stress, expected):
    np.testing.assert_allclose(ramberg_osgood_monotone.plastic_strain(stress), expected, rtol=1e-5)


@pytest.mark.parametrize('stress, expected', map(tuple, parametrization_data_monotone_plastic))
def test_rambgood_plastic_strain_array(ramberg_osgood_monotone, stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood_monotone.plastic_strain(stress), expected, significant=5)


parametrization_data_monotone_tang_compl = np.array([
    [0., 1./2],
    [1., 5./9],
    [2., 11./18],
    [4., 13./18],
    [-1., 5./9],
    [-2., 11./18],
    [-4., 13./18]
])


@pytest.mark.parametrize('stress, expected', map(tuple, parametrization_data_monotone_tang_compl))
def test_rambgood_tangential_compliance_scalar(ramberg_osgood_monotone, stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood_monotone.tangential_compliance(stress), expected, significant=5)


@pytest.mark.parametrize('stress, expected', [
    (parametrization_data_monotone_tang_compl[:, 0], parametrization_data_monotone_tang_compl[:, 1])
])
def test_rambgood_tangential_compliance_array(ramberg_osgood_monotone, stress, expected):
    np.testing.assert_allclose(ramberg_osgood_monotone.tangential_compliance(stress), expected, rtol=1e-5)


@pytest.mark.parametrize('stress, expected', map(tuple, parametrization_data_monotone_tang_compl))
def test_rambgood_tangential_modulus_scalar(ramberg_osgood_monotone, stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood_monotone.tangential_modulus(stress), 1./expected, significant=5)


@pytest.mark.parametrize('stress, expected', [
    (parametrization_data_monotone_tang_compl[:, 0], parametrization_data_monotone_tang_compl[:, 1])
])
def test_rambgood_tangential_modulus_array(ramberg_osgood_monotone, stress, expected):
    np.testing.assert_allclose(ramberg_osgood_monotone.tangential_modulus(stress), 1./expected, rtol=1e-5)


@pytest.fixture
def ramberg_osgood_cyclic():
    return RambergOsgood(E=2., K=3., n=0.5)


parametrization_data_delta = np.array([
    [0.0, 0.0],
    [1.0, 10./18.],
    [2.0, 11./9.],
    [-1.0, -10./18.],
    [-2.0, -11./9.]
])


@pytest.mark.parametrize('delta_stress, expected', map(tuple, parametrization_data_delta))
def test_rambgood_delta_strain_scalar(ramberg_osgood_cyclic, delta_stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood_cyclic.delta_strain(delta_stress), expected, significant=5)


@pytest.mark.parametrize('expected, delta_strain', map(tuple, parametrization_data_delta))
def test_rambgood_delta_stress_scalar(ramberg_osgood_cyclic, expected, delta_strain):
    np.testing.assert_approx_equal(ramberg_osgood_cyclic.delta_stress(delta_strain), expected, significant=5)


@pytest.mark.parametrize('delta_stress, expected', [(parametrization_data_delta[:, 0], parametrization_data_delta[:, 1])])
def test_rambgood_delta_strain_array(ramberg_osgood_cyclic, delta_stress, expected):
    np.testing.assert_allclose(ramberg_osgood_cyclic.delta_strain(delta_stress), expected, rtol=1e-5)


@pytest.mark.parametrize('expected, delta_strain', [(parametrization_data_delta[:, 0], parametrization_data_delta[:, 1])])
def test_rambgood_delta_stress_array(ramberg_osgood_cyclic, expected, delta_strain):
    np.testing.assert_allclose(ramberg_osgood_cyclic.delta_stress(delta_strain), expected, rtol=1e-5)


@pytest.fixture
def ramberg_osgood():
    return RambergOsgood(210.5e9, 1078e6, 0.133)


def test_rambgood_char_init(ramberg_osgood):
    assert ramberg_osgood._E == 210.5e9
    assert ramberg_osgood._K == 1078e6
    assert ramberg_osgood._n == 0.133


def test_rambgood_properties_real(ramberg_osgood):
    assert ramberg_osgood.E == ramberg_osgood._E
    assert ramberg_osgood.K == ramberg_osgood._K
    assert ramberg_osgood.n == ramberg_osgood._n


parametrization_data_monotone_real = np.array([
    [0.0, 0.0],
    [700e6, 0.042236],
    [1000e6, 0.57327],
    [-700e6, -0.042236],
    [-1000e6, -0.57327]
])


@pytest.mark.parametrize('stress, expected', map(tuple, parametrization_data_monotone_real))
def test_rambgood_char_strain_scalar(ramberg_osgood, stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood.strain(stress), expected, significant=5)


@pytest.mark.parametrize('expected, strain', map(tuple, parametrization_data_monotone_real))
def test_rambgood_char_stress_scalar(ramberg_osgood, expected, strain):
    np.testing.assert_approx_equal(ramberg_osgood.stress(strain), expected, significant=5)


@pytest.mark.parametrize('stress, expected', [(parametrization_data_monotone_real[:, 0], parametrization_data_monotone_real[:, 1])])
def test_rambgood_char_strain_array(ramberg_osgood, stress, expected):
    np.testing.assert_allclose(ramberg_osgood.strain(stress), expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [(parametrization_data_monotone_real[:, 0], parametrization_data_monotone_real[:, 1])])
def test_rambgood_char_stress_array(ramberg_osgood, expected, strain):
    np.testing.assert_allclose(ramberg_osgood.stress(strain), expected, rtol=1e-5)


parametrization_data_delta_real = np.array([
    [0.0, 0.0],
    [1000e6, 0.01095061],
    [1100e6, 0.01792016],
    [-1000e6, -0.01095061],
    [-1100e6, -0.01792016]
])


@pytest.mark.parametrize('delta_stress, expected', map(tuple, parametrization_data_delta_real))
def test_rambgood_char_delta_strain_scalar(ramberg_osgood, delta_stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood.delta_strain(delta_stress), expected, significant=5)


@pytest.mark.parametrize('expected, delta_strain', map(tuple, parametrization_data_delta_real))
def test_rambgood_char_delta_stress_scalar(ramberg_osgood, expected, delta_strain):
    np.testing.assert_approx_equal(ramberg_osgood.delta_stress(delta_strain), expected, significant=5)


@pytest.mark.parametrize('delta_stress, expected', [(parametrization_data_delta_real[:, 0], parametrization_data_delta_real[:, 1])])
def test_rambgood_char_delta_strain_array(ramberg_osgood, delta_stress, expected):
    np.testing.assert_allclose(ramberg_osgood.delta_strain(delta_stress), expected, rtol=1e-5)


parametrization_data_monotone_tang_compl_real = np.array([
    [0., 1/(210.5e9)],
    [700e6, 4.2269651e-10],
    [1000e6, 4.2793411e-09],
    [-700e6, 4.2269651e-10],
    [-1000e6, 4.2793411e-09],
])


@pytest.mark.parametrize('stress, expected', map(tuple, parametrization_data_monotone_tang_compl_real))
def test_rambgood_tangential_compliance_real_scalar(ramberg_osgood, stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood.tangential_compliance(stress), expected, significant=5)


@pytest.mark.parametrize('stress, expected', [
    (parametrization_data_monotone_tang_compl_real[:, 0], parametrization_data_monotone_tang_compl_real[:, 1])
])
def test_rambgood_tangential_compliance_real_array(ramberg_osgood, stress, expected):
    np.testing.assert_allclose(ramberg_osgood.tangential_compliance(stress), expected, rtol=1e-5)


@pytest.mark.parametrize('stress, expected', map(tuple, parametrization_data_monotone_tang_compl_real))
def test_rambgood_tangential_modulus_real_scalar(ramberg_osgood, stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood.tangential_modulus(stress), 1./expected, significant=5)


@pytest.mark.parametrize('stress, expected', [
    (parametrization_data_monotone_tang_compl_real[:, 0], parametrization_data_monotone_tang_compl_real[:, 1])
])
def test_rambgood_tangential_modulus_real_array(ramberg_osgood, stress, expected):
    np.testing.assert_allclose(ramberg_osgood.tangential_modulus(stress), 1./expected, rtol=1e-5)


@pytest.mark.parametrize('stress, max_stress, expected', [
    (-700e6, 700e6, -0.042236),
    (-1000e6, 1000e6, -0.57327),
    (0.0, 1000e6, 0.562320)
])
def test_rambgood_char_lower_hysteresis_scalar(ramberg_osgood, stress, max_stress, expected):
    np.testing.assert_approx_equal(ramberg_osgood.lower_hysteresis(stress, max_stress),
                                   expected, significant=5)


@pytest.mark.parametrize('stress, max_stress, expected', [
    (np.array([-1000e6, 0.0]), 1000e6, np.array([-0.57327, 0.562320])),
])
def test_rambgood_char_lower_hysteresis_array(ramberg_osgood, stress, max_stress, expected):
    np.testing.assert_allclose(ramberg_osgood.lower_hysteresis(stress, max_stress),
                               expected, rtol=1e-5)


def test_ramgood_lower_hysteresis_above_max_stress_scalar(ramberg_osgood):
    with pytest.raises(ValueError, match=r'Value for \'stress\' must not be higher than \'max_stress\'.'):
        ramberg_osgood.lower_hysteresis(200., 100.)


def test_ramgood_lower_hysteresis_above_max_stress_array(ramberg_osgood):
    with pytest.raises(ValueError, match=r'Value for \'stress\' must not be higher than \'max_stress\'.'):
        ramberg_osgood.lower_hysteresis([50., 200.], 100.)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pylife.utils.functions as F


def test_rossow():
    assert F.rossow_cumfreqs(1) == 0.5
    for N in np.random.randint(1, high=100, size=100):
        cf = F.rossow_cumfreqs(N)
        np.testing.assert_approx_equal(cf.sum(), N/2.0)
        for i in range(N):
            assert cf[i] + cf[N-i-1] == 1.0


def test_scattering_range_2_std():
    np.testing.assert_allclose(F.scattering_range_to_std(1.25), 0.0378096, rtol=1e-5)


def test_std_2_scattering_range():
    np.testing.assert_allclose(F.std_to_scattering_range(0.0378096), 1.25, rtol=1e-5)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Small helper functions for fatigue analysis

"""

__author__ = "Cedric Philip Wagner"
__maintainer__ = "Johannes Mueller"


import pandas as pd
import numpy as np

import pylife.stress.collective as CL

@pd.api.extensions.register_series_accessor('solidity')
class SolidityAccessor(CL.LoadHistogram):

    def haibach(self, k):
        return haibach(self, k)

    def fkm(self, k):
        return fkm(self, k)


def haibach(collective, k):
    """Compute solidity according to Haibach

    Refer to:
    Haibach - Betriebsfestigkeit - 3. Auflage (2005) - S.271

    Parameters
    ----------
    collective : np.ndarray
        numpy array of shape (:, 2) where ":" depends on the number of classes
        defined for the rainflow counting

            1. column: class values in ascending order

            2. column: accumulated number of cycles first entry is the total
               number of cycles then in a descending manner till the number of
               cycles of the highest stress class

    k : float
        slope of the S/N curve

    Returns
    -------
    V : np.ndarray (1,)
        Völligkeitswert (solidity)

    """

    S = collective.amplitude
    hi = collective.cycles

    xi = S / S[hi > 0].max()
    V = np.sum((hi * (xi**k)) / hi.sum())

    return V


def fkm(collective, k):
    """Compute solidity according to the FKM guideline (2012)

    Refer to:
    FKM-Richtlinie - 6. Auflage (2012) - S.58 - Gl. (2.4.55) +  Gl. (2.4.55)

    Parameters
    ----------
    collective : np.ndarray
        numpy array of shape (:, 2) where ":" depends on the number of classes
        defined for the rainflow counting

            1. column: class values in ascending order

            2. column: accumulated number of cycles first entry is the total
               number of cycles then in a descending manner till the number of
               cycles of the highest stress class k : float slope of the S/N
               curve

    Returns
    -------
    V : np.ndarray
        Völligkeitswert (solidity)

    """

    V_haibach = haibach(collective, k)
    V = V_haibach**(1./k)

    return V



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np
import numpy.random as rnd
from scipy.stats import norm

import pylife.strength.failure_probability as FP

import numpy.testing as testing


def test_simple_load():
    pf = FP.FailureProbability(100., 5.)
    r = pf.pf_simple_load(100.)
    assert r == 0.5


def test_norm_load():
    pf = FP.FailureProbability(100., 5.)
    r = pf.pf_norm_load(100., 1e-32)
    testing.assert_almost_equal(r, 0.5)


def test_arbitrary_load():
    strength_median = 100.
    strength_std = 5.

    load_median = 70.
    load_std = 13.

    pf = FP.FailureProbability(strength_median, strength_std)
    expected = pf.pf_norm_load(load_median, load_std)

    lower = np.log10(load_median) - 13*load_std
    upper = np.log10(load_median) + 7*load_std
    load = lower + np.sort(rnd.rand(1500)*(upper-lower))

    pdf = norm.pdf(load, loc=np.log10(load_median), scale=load_std)

    r = pf.pf_arbitrary_load(load, pdf)

    testing.assert_almost_equal(r, expected, decimal=3)


def test_arbitrary_load_incompatible_dimensions():
    pf = FP.FailureProbability(1.0, 0.1)
    with pytest.raises(ValueError, match=r'Load values and pdf must have same dimensions.'):
        pf.pf_arbitrary_load(np.array([1, 2, 3]), np.array([0.1, 0.2]))



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np
import scipy.stats as stats

from pylife.utils.functions import scattering_range_to_std

from pylife import PylifeSignal


@pd.api.extensions.register_series_accessor('woehler')
@pd.api.extensions.register_dataframe_accessor('woehler')
class WoehlerCurve(PylifeSignal):
    """A PylifeSignal accessor for Wöhler Curve data.

    Wöhler Curve (aka SN-curve) determines after how many load cycles at a
    certain load amplitude the component is expected to fail.

    The signal has the following mandatory keys:

    * ``k_1`` : The slope of the Wöhler Curve
    * ``ND`` : The cycle number of the endurance limit
    * ``SD`` : The load level of the endurance limit

    The ``_50`` suffixes imply that the values are valid for a 50% probability
    of failure.

    There are the following optional keys:

    * ``k_2`` : The slope of the Wöhler Curve below the endurance limit
                If the key is missing it is assumed to be infinity, i.e. perfect endurance
    * ``TN`` : The scatter in cycle direction, (N_90/N_10)
               If the key is missing it is assumed to be 1.0 – i.e. no scatter –
               or calculated from ``TS`` if given.
    * ``TS`` : The scatter in load direction, (SD_90/SD_10)
               If the key is missing it is assumed to be 1.0 – i.e. no scatter –
               or calculated from ``TN`` if given.
    """

    def __init__(self, pandas_obj):
        self._obj = pandas_obj.copy()
        self._validate()

    def _validate(self):
        self.fail_if_key_missing(['k_1', 'ND', 'SD'])
        self._k_2 = self._obj.get('k_2', np.inf)

        self._TN = self._obj.get('TN', None)
        self._TS = self._obj.get('TS', None)

        if self._TN is None and self._TS is None:
            self._TN = 1.0
            self._TS = 1.0
        elif self._TS is None:
            self._TS = np.power(self._TN, 1./self._obj.k_1)
        elif self._TN is None:
            self._TN = np.power(self._TS, self._obj.k_1)

        self._failure_probability = self._obj.get('failure_probability', 0.5)

        self._obj['k_2'] = self._k_2
        self._obj['TN'] = self._TN
        self._obj['TS'] = self._TS
        self._obj['failure_probability'] = self._failure_probability

    @property
    def SD(self):
        return self._obj.SD

    @property
    def ND(self):
        return self._obj.ND

    @property
    def k_1(self):
        """The second Wöhler slope."""
        return self._obj.k_1

    @property
    def k_2(self):
        """The second Wöhler slope."""
        return self._obj.k_2

    @property
    def TN(self):
        """The load direction scatter value TN."""
        return self._obj.TN

    @property
    def TS(self):
        """The load direction scatter value TS."""
        return self._obj.TS

    @property
    def failure_probability(self):
        return self._failure_probability

    def transform_to_failure_probability(self, failure_probability):
        failure_probability = np.asarray(failure_probability, dtype=np.float64)

        failure_probability, obj = self.broadcast(failure_probability)

        native_ppf = stats.norm.ppf(obj.failure_probability)
        goal_ppf = stats.norm.ppf(failure_probability)

        SD = np.asarray(obj.SD / 10**((native_ppf-goal_ppf)*scattering_range_to_std(obj.TS)))
        ND = np.asarray(obj.ND / 10**((native_ppf-goal_ppf)*scattering_range_to_std(obj.TN)))
        ND.flags.writeable = True
        ND[SD != 0] *= np.power(SD[SD != 0]/obj.SD, -obj.k_1)

        transformed = obj.copy()
        transformed['SD'] = SD
        transformed['ND'] = ND
        transformed['failure_probability'] = failure_probability

        return WoehlerCurve(transformed)

    def miner_original(self):
        """Set k_2 to inf according Miner Original method (k_2 = inf).

        Returns
        -------
        modified copy of self
        """
        new = self._obj.copy()
        new['k_2'] =  np.inf
        return self.__class__(new)

    def miner_elementary(self):
        """Set k_2 to k_1 according Miner Elementary method (k_2 = k_1).

        Returns
        -------
        modified copy of self
        """
        new = self._obj.copy()
        new['k_2'] =  self._obj.k_1
        return self.__class__(new)

    def miner_haibach(self):
        """Set k_2 to value according Miner Haibach method (k_2 = 2 * k_1 - 1).

        Returns
        -------
        modified copy of self
        """
        new = self._obj.copy()
        new['k_2'] = 2. * self._obj.k_1 - 1.
        return self.__class__(new)

    def cycles(self, load, failure_probability=0.5):
        """Calculate the cycles numbers from loads.

        Parameters
        ----------
        load : array_like
            The load levels for which the corresponding cycle numbers are to be calculated.
        failure_probability : float, optional
            The failure probability with which the component should fail when
            charged with `load` for the calculated cycle numbers. Default 0.5

        Returns
        -------
        cycles : numpy.ndarray
            The cycle numbers at which the component fails for the given `load` values


        Notes
        -----
        By default the calculation is performed according to the Basquin
        equation using :meth:`basquin_cycles`.  Derived classes can choose to
        override this in order to implement a different fatigue law.
        """
        return self.basquin_cycles(load, failure_probability)

    def load(self, cycles, failure_probability=0.5):
        """Calculate the load values from loads.

        Parameters
        ----------
        cycles : array_like
            The cycle numbers for which the corresponding load levels are to be calculated.
        failure_probability : float, optional
            The failure probability with which the component should fail when
            charged with `load` for the calculated cycle numbers. Default 0.5

        Returns
        -------
        cycles : numpy.ndarray
            The cycle numbers at which the component fails for the given `load` values

        Notes
        -----
        By default the calculation is performed according to the Basquin
        equation using :meth:`basquin_cycles`.  Derived classes can choose to
        override this in order to implement a different fatigue law.
        """
        return self.basquin_load(cycles, failure_probability)

    def basquin_cycles(self, load, failure_probability=0.5):
        """Calculate the cycles numbers from loads according to the Basquin equation.

        Parameters
        ----------
        load : array_like
            The load levels for which the corresponding cycle numbers are to be calculated.
        failure_probability : float, optional
            The failure probability with which the component should fail when
            charged with `load` for the calculated cycle numbers. Default 0.5

        Returns
        -------
        cycles : numpy.ndarray
            The cycle numbers at which the component fails for the given `load` values
        """
        def ensure_float_to_prevent_int_overflow(load):
            if isinstance(load, pd.Series):
                return pd.Series(load, dtype=np.float64)
            return np.asarray(load, dtype=np.float64)

        transformed = self.transform_to_failure_probability(failure_probability)

        load = ensure_float_to_prevent_int_overflow(load)
        ld, wc = transformed.broadcast(load)
        cycles = np.full_like(ld, np.inf)

        k = self._make_k(ld, wc.SD, wc)
        in_limit = np.isfinite(k)
        cycles[in_limit] = wc.ND[in_limit] * np.power(ld[in_limit]/wc.SD[in_limit], -k[in_limit])

        if not isinstance(load, pd.Series):
            return cycles
        return pd.Series(cycles, index=ld.index)

    def basquin_load(self, cycles, failure_probability=0.5):
        """Calculate the load values from loads according to the Basquin equation.

        Parameters
        ----------
        cycles : array_like
            The cycle numbers for which the corresponding load levels are to be calculated.
        failure_probability : float, optional
            The failure probability with which the component should fail when
            charged with `load` for the calculated cycle numbers. Default 0.5

        Returns
        -------
        cycles : numpy.ndarray
            The cycle numbers at which the component fails for the given `load` values
        """
        transformed = self.transform_to_failure_probability(failure_probability)

        cyc, wc = transformed.broadcast(cycles)
        load = np.asarray(wc.SD).copy()

        k = self._make_k(-cyc, -wc.ND, wc)
        in_limit = np.isfinite(k)
        load[in_limit] = wc.SD[in_limit] * np.power(cyc[in_limit]/wc.ND[in_limit], -1./k[in_limit])

        if not isinstance(cycles, pd.Series):
            return load
        return pd.Series(load, index=cyc.index)

    def _make_k(self, src, ref, wc):
        k = np.asarray(wc.k_1).copy()
        k_2 = np.asarray(wc.k_2)

        below_limit = np.asarray(src < ref)
        if k.shape == ():
            k = np.full_like(src, k, dtype=np.double)
            k_2 = np.full_like(src, k_2, dtype=np.double)

        k[below_limit] = k_2[below_limit]
        return k



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import pytest


def test_clean_timeseries_no_tsfresh():
    sys.modules['tsfresh'] = None

    sys.modules.pop('pylife.stress.timesignal', None)

    import pylife.stress.timesignal as pts

    with pytest.raises(ImportError, match=r"pip install pylife\[tsfresh\]"):
        pts.clean_timeseries(None, None)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import os
import sys
import time
import pickle
import subprocess as sp
import shutil

import threading as THR
import queue as QU

import numpy as np
import pandas as pd


class OdbServerError(Exception):
    """Raised when the ODB Server launch fails."""

    pass


class OdbClient:
    """The interface class to access data from odb files provided by the odbserver.

    Parameters
    ----------
    odb_file : string
        The path to the odb file

    abaqus_bin : string, optional
        The path to the abaqus *binary* (not a .bat or shell script).
        Guessed if not given.

    python_env_path : string, optional
        The path to the python2 environmnent to be used by the odbserver.
        Guessed if not given.

    Examples
    --------
    Instantiating and querying instance names

    >>> import odbclient as CL
    >>> client = CL.OdbClient("some_file.odb")
    >>> client.instance_names()
    ['PART-1-1']

    Querying node coordinates

    >>> client.node_coordinates('PART-1-1')
                x     y     z
    node_id
    1       -30.0  15.0  10.0
    2       -30.0  25.0  10.0
    3       -30.0  15.0   0.0
    ...

    Querying step names

    >>> client.step_names()
    ['Load']

    Querying frames of a step

    >>> client.frame_ids('Load')
    [0, 1]

    Querying variable names of a frame and step

    >>> client.variable_names('Load', 1)
    ['CF', 'COORD', 'E', 'EVOL', 'IVOL', 'RF', 'S', 'U']

    Querying variable data of an instance, frame and step

    >>> client.variable('S', 'PART-1-1', 'Load', 1)
                              S11        S22  ...       S13       S23
    node_id element_id                        ...
    5       1          -38.617779   2.705118  ... -3.578981  1.355571
    7       1          -38.617779   2.705118  ...  3.578981 -1.355571
    3       1          -50.749348 -21.749729  ... -7.597347 -0.000003
    1       1          -50.749348 -21.749729  ...  7.597347  0.000003
    6       1           38.643414  -2.588303  ...  3.522046  1.446851
    ...                       ...        ...  ...       ...       ...
    54      4            7.353698  -3.177251  ...  1.775653 -2.608372
    56      4           -6.695759 -17.656754  ...  0.217049 -3.040078
    55      4           -6.695759 -17.656754  ... -0.217049  3.040078
    47      4           -0.226473   1.787100  ...  0.967435 -0.671089
    48      4           -0.226473   1.787100  ... -0.967435  0.671089


    Oftentimes it is desirable to have the node coordinates and multiple field
    variables in one dataframe.  This can be easily achieved by
    :meth:`~pandas.DataFrame.join` operations.

    >>> node_coordinates = client.node_coordinates('PART-1-1')
    >>> stress = client.variable('S', 'PART-1-1', 'Load', 1)
    >>> strain = client.variable('E', 'PART-1-1', 'Load', 1)
    >>> node_coordinates.join(stress).join(strain)
                           x     y     z  ...           E12           E13           E23
    node_id element_id                    ...
    5       1          -20.0  15.0  10.0  ... -2.741873e-11 -4.652675e-11  1.762242e-11
    7       1          -20.0  15.0   0.0  ... -2.741873e-11  4.652675e-11 -1.762242e-11
    3       1          -30.0  15.0   0.0  ... -2.599339e-11 -9.876550e-11 -3.946581e-17
    1       1          -30.0  15.0  10.0  ... -2.599339e-11  9.876550e-11  3.946581e-17
    6       1          -20.0  25.0  10.0  ... -2.689760e-11  4.578660e-11  1.880906e-11
    ...                  ...   ...   ...  ...           ...           ...           ...
    54      4            5.0  25.0  10.0  ... -6.076223e-11  2.308349e-11 -3.390884e-11
    56      4           10.0  20.0  10.0  ... -5.091068e-11  2.821631e-12 -3.952102e-11
    55      4           10.0  20.0   0.0  ... -5.091068e-11 -2.821631e-12  3.952102e-11
    47      4            0.0  20.0   0.0  ... -5.129363e-11  1.257666e-11 -8.724152e-12
    48      4            0.0  20.0  10.0  ... -5.129363e-11 -1.257666e-11  8.724152e-12
    """

    def __init__(self, odb_file, abaqus_bin=None, python_env_path=None):
        self._proc = None
        env = os.environ
        env['PYTHONPATH'] = _guess_pythonpath(python_env_path)

        lock_file_exists = os.path.isfile(os.path.splitext(odb_file)[0] + '.lck')

        abaqus_bin = abaqus_bin or _guess_abaqus_bin()

        self._proc = sp.Popen([abaqus_bin, 'python', '-m', 'odbserver', odb_file],
                              stdout=sp.PIPE, stdin=sp.PIPE, stderr=sp.PIPE,
                              env=env)

        if lock_file_exists:
            self._gulp_lock_file_warning()

        self._wait_for_server_ready_sign()

    def _gulp_lock_file_warning(self):
        self._proc.stdout.readline()
        self._proc.stdout.readline()

    def _wait_for_server_ready_sign(self):
        def wait_for_input(stdout, queue):
            sign = stdout.read(5)
            queue.put(sign)

        queue = QU.Queue()
        thread = THR.Thread(target=wait_for_input, args=(self._proc.stdout, queue))
        thread.daemon = True
        thread.start()

        while True:
            self._check_if_process_still_alive()
            try:
                sign = queue.get_nowait()
            except QU.Empty:
                time.sleep(1)
            else:
                if sign != b'ready':
                    raise OdbServerError("Expected ready sign from server, received %s" % sign)
                return

    def instance_names(self):
        """Query the instance names from the odbserver.

        Returns
        -------
        instance_names : list of string
            The names of the instances.
        """
        return _ascii(_decode, self._query('get_instances'))

    def node_coordinates(self, instance_name, nset_name=''):
        """Query the node coordinates of an instance.

        Parameters
        ----------
        instance_name : string
            The name of the instance to be queried
        nset_name : string, optional
            A name of a node set of the instance that the query is to be limited to.

        Returns
        -------
        node_coords : :class:`pandas.DataFrame`
            The node list as a pandas data frame without connectivity.
            The columns are named ``x``, ``y`` and ``z``.
        """
        self._fail_if_instance_invalid(instance_name)
        index, node_data = self._query('get_nodes', (instance_name, nset_name))
        return pd.DataFrame(data=node_data, columns=['x', 'y', 'z'],
                            index=pd.Index(index, name='node_id', dtype=np.int64))

    def element_connectivity(self, instance_name, elset_name=''):
        """Query the element connectivity of an instance.

        Parameters
        ----------
        instance_name : string
            The name of the instance to be queried
        elset_name : string, optional
            A name of an element set of the instance that the query is to be limited to.

        Returns
        -------
        connectivity : :class:`pandas.DataFrame`
            The connectivity as a :class:`pandas.DataFrame`.
            For every element there is list of node ids that the element is connected to.

        """
        index, connectivity = self._query('get_connectivity', (instance_name, elset_name))
        return pd.DataFrame({'connectivity': connectivity},
                            index=pd.Index(index, name='element_id', dtype=np.int64))

    def nset_names(self, instance_name=''):
        """Query the available node set names.

        Parameters
        ----------
        instance_name : string, optional
            The name of the instance the node sets are queried from. If not given the
            node sets of all instances are returned.

        Returns
        -------
        instance_names : list of strings
            The names of the instances
        """
        self._fail_if_instance_invalid(instance_name)
        return _ascii(_decode, self._query('get_node_sets', instance_name))

    def node_ids(self, nset_name, instance_name=''):
        """Query the node ids of a certain node set.

        Parameters
        ----------
        nset_name : string
            The name of the node set
        instance_name : string, optional
            The name of the instance the node set is to be taken from. If not given
            node sets from all instances are considered.

        Returns
        -------
        node_ids : :class:`pandas.Index`
            The node ids as :class:`pandas.Index`
        """
        node_ids = self._query('get_node_set', (instance_name, nset_name))
        return pd.Index(node_ids, name='node_id', dtype=np.int64)

    def elset_names(self, instance_name=''):
        """Query the available element set names.

        Parameters
        ----------
        instance_name : string, optional
            The name of the instance the element sets are queried from. If not given the
            element sets of all instances are returned.

        Returns
        -------
        instance_names : list of strings
            The names of the instances
        """
        self._fail_if_instance_invalid(instance_name)
        return _ascii(_decode, self._query('get_element_sets', instance_name))

    def element_ids(self, elset_name, instance_name=''):
        """Query the element ids of a certain element set.

        Parameters
        ----------
        elset_name : string
            The name of the element set
        instance_name : string, optional
            The name of the instance the element set is to be taken from. If not given
            element sets from all instances are considered.

        Returns
        -------
        element_ids : :class:`pandas.Index`
            The element ids as :class:`pandas.Index`
        """
        element_ids = self._query('get_element_set', (instance_name, elset_name))
        return pd.Index(element_ids, name='element_id', dtype=np.int64)

    def step_names(self):
        """Query the step names from the odb file.

        Returns
        -------
        step_names : list of string
            The names of all the steps stored in the odb file.
        """
        return _ascii(_decode, self._query('get_steps'))

    def frame_ids(self, step_name):
        """Query the frames of a given step.

        Parameters
        ----------
        step_name : string
            The name of the step

        Returns
        -------
        step_name : list of ints
            The name of the step the frame ids are expected in.
        """
        return self._query('get_frames', step_name)

    def variable_names(self, step_name, frame_id):
        """Query the variable names of a certain step and frame.

        Parameters
        ----------
        step_name : string
            The name of the step
        frame_id : int
            The index of the frame

        Returns
        -------
        variable_names : list of string
            The names of the variables
        """
        return _ascii(_decode, self._query('get_variable_names', (step_name, frame_id)))

    def variable(self, variable_name, instance_name, step_name, frame_id, nset_name='', elset_name='', position=None):
        """Read field variable data.

        Parameters
        ----------
        variable_name : string
            The name of the variable.
        instance_name : string
            The name of the instance.
        step_name : string
            The name of the step
        frame_id : int
            The index of the frame
        nset_name : string, optional
            The name of the node set to be queried. If not given, the whole instance
        elnset_name : string, optional
            The name of the element set to be queried. If not given, the whole instance
        position : string, optional
            Position within element. Terminology as in Abaqus .inp file:
            ``INTEGRATION POINTS``, ``CENTROIDAL``, ``WHOLE ELEMENT``, ``NODES``,
            ``FACES``, ``AVERAGED AT NODES``

            If not given the native position is taken, except for ``INTEGRATION_POINTS``
            The ``ELEMENT_NODAL`` position is used.
        """
        response = self._query('get_variable', (instance_name, step_name, frame_id, variable_name, nset_name, elset_name, position))
        (labels, index_labels, index_data, values) = response

        index_labels = _ascii(_decode, index_labels)
        if len(index_labels) > 1:
            index = pd.DataFrame(index_data, columns=index_labels, dtype=np.int64).set_index(index_labels).index
        else:
            index = pd.Index(index_data[:, 0], name=index_labels[0], dtype=np.int64)

        column_names = _ascii(_decode, labels)
        return pd.DataFrame(values, index=index, columns=column_names)

    def history_regions(self, step_name):
         """Query the history Regions of a given step.

         Parameters
         ----------
         step_name : string
             The name of the step

         Returns
         -------
         historyRegions : list of strings
             The name of history regions, which are in the required step.
         """
         return self._query('get_history_regions', step_name)

    def history_outputs(self, step_name, history_region_name):
         """Query the history Outputs of a given step in a given history region.

         Parameters
         ----------
         step_name : string
             The name of the step

         history_region_name: string
             The name of the history region

         Returns
         -------
         historyOutputs : list of strings
             The name of the history outputs, which are in the required step and under the required history region
         """
         hisoutputs = self._query("get_history_outputs", (step_name, history_region_name))

         return hisoutputs


    def history_output_values(self, step_name, history_region_name, historyoutput_name):
         """Query the history Regions of a given step.

         Parameters
         ----------
         step_name : string
             The name of the step

         Returns
         -------
         historyRegions : list of strings
             The name of the step the history regions are in.
         """
         hisoutput_valuesx, hisoutput_valuesy = self._query("get_history_output_values", (step_name, history_region_name, historyoutput_name))
         history_region_description = self._query("get_history_region_description", (step_name, history_region_name))
         historyoutput_data = pd.Series(hisoutput_valuesy, index = hisoutput_valuesx, name = history_region_description + ": " + historyoutput_name)

         return historyoutput_data

    def history_region_description(self, step_name, history_region_name):
         """Query the description of a history Regions of a given step.

         Parameters
         ----------
         step_name : string
             The name of the step
         history_region_name: string
             The name of the history region

         Returns
         -------
         historyRegion_description : list of strings
             The description of the history region.
         """
         history_region_description = self._query("get_history_region_description", (step_name, history_region_name))
         return history_region_description

    def history_info(self):
        """Query all the information about the history outputs in a given odb.
         Returns
         -------
         dictionary : ldictionary which contains history information
         """
        dictionary = _decode(self._query("get_history_info"))
        return dictionary

    def _query(self, command, args=None):
        args = _ascii(_encode, args)
        self._send_command(command, args)
        self._check_if_process_still_alive()
        array_num, pickle_data = self._parse_response()

        if isinstance(pickle_data, Exception):
            raise pickle_data

        if array_num == 0:
            return _ascii(_decode, pickle_data)

        numpy_arrays = [np.lib.format.read_array(self.proc.stdout) for _ in range(array_num)]

        return _ascii(_decode, pickle_data), numpy_arrays

    def _send_command(self, command, args=None):
        self._check_if_process_still_alive()
        pickle.dump((command, args), self._proc.stdin, protocol=2)
        self._proc.stdin.flush()

    def _parse_response(self):
        pickle_data = b''
        while True:
            line = self._proc.stdout.readline().rstrip() + b'\n'
            pickle_data += line
            if line == b'.\n':
                break
        return pickle.loads(pickle_data, encoding='bytes')

    def __del__(self):
        if self._proc is not None:
            self._send_command('QUIT')
            time.sleep(1)

    def _check_if_process_still_alive(self):
        if self._proc.poll() is not None:
            _, error_message = self._proc.communicate()
            self._proc = None

            raise OdbServerError(error_message.decode('ascii'))

    def _fail_if_instance_invalid(self, instance_name):
        if instance_name not in self.instance_names() and instance_name != '':
            raise KeyError("Invalid instance name '%s'." % instance_name)


def _ascii(fcn, args):
    if isinstance(args, list):
        return [_ascii(fcn, arg) for arg in args]

    if isinstance(args, tuple):
        return tuple(_ascii(fcn, arg) for arg in args)

    return fcn(args)


def _encode(arg):
    return arg.encode('ascii') if isinstance(arg, str) else arg


def _decode(arg):
    if isinstance(arg, bytes):
        return arg.decode('ascii')
    if isinstance(arg, dict):
        return {_decode(key): _decode(value) for key, value in arg.items()}
    if isinstance(arg, list):
        return [_decode(element) for element in arg]
    return arg

def _guess_abaqus_bin():
    if sys.platform == 'win32':
        return _guess_abaqus_bin_windows()
    return shutil.which('abaqus')


def _guess_abaqus_bin_windows():
    guesses = [
        r"C:/Program Files/SIMULIA/2018/AbaqusCAE/win_b64/code/bin/ABQLauncher.exe",
        r"C:/Program Files/SIMULIA/2020/EstProducts/win_b64/code/bin/ABQLauncher.exe",
        r"C:/Program Files/SIMULIA/2020/Products/win_b64/code/bin/ABQLauncher.exe",
        r"C:/Program Files/SIMULIA/2021/EstProducts/win_b64/code/bin/ABQLauncher.exe",
    ]
    for guess in guesses:
        if os.path.exists(guess):
            return guess
    raise OSError("Could not guess abaqus binary path! Please submit as abaqus_bin parameter!")


def _guess_pythonpath(python_env_path):
    python_env_path = _guess_python_env_path(python_env_path)
    if python_env_path is None:
        raise OSError("No odbserver environment found.\n"
                      "Please see https://github.com/boschresearch/pylife/blob/develop/tools/odbserver/README.md")
    if sys.platform == 'win32':
        return os.path.join(python_env_path, 'lib', 'site-packages')
    return os.path.join(python_env_path, 'lib', 'python2.7', 'site-packages')


def _guess_python_env_path(python_env_path):
    cand = python_env_path or os.path.join(os.environ['HOME'], '.conda', 'envs', 'odbserver')
    if os.path.exists(cand):
        return cand
    return None
